/******************************************************************************/
/*                                                                            */
/*                          X r d D P M O f s . c c                           */
/*                                                                            */
/* DPM handling and various by David Smith, CERN                              */            
/* Other parts:                                                               */
/*                                                                            */
/* (c) 2006 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

static const char XrdDPMOfsCVSID[] = "$Id: XrdDPMOfs.cc 3618 2010-06-08 14:21:17Z dhsmith $";

#include "XrdDPMOfs.hh"
#include "XrdDPMOfsTrace.hh"
#include "XrdDPMVersion.hh"

#include "XrdSys/XrdSysError.hh"
#include "XrdOuc/XrdOucTrace.hh"
#include "XrdOss/XrdOssApi.hh"
#include "XrdOuc/XrdOucString.hh"
#include "XrdSys/XrdSysPthread.hh"
#include "XrdOuc/XrdOucName2Name.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdSec/XrdSecEntity.hh"
#include "XrdOuc/XrdOuca2x.hh"

#include <pwd.h>
#include <sys/types.h>
#include <vector>
#include <map>
#include <string>
#include <iostream>
#include <sstream>

#include <dlfcn.h>
#ifndef __macos__
#include <link.h>
#endif

// no prototype for these two elsewhere
//
extern "C" int Cdomainname(char *name, int namele);
extern "C" int Cns_client_resetAuthorizationId();

#include "dpns_api.h"
#include "dpm_api.h"
#include "serrno.h"

/******************************************************************************/
/*                  E r r o r   R o u t i n g   O b j e c t                   */
/******************************************************************************/

static XrdSysError      DPMOfsEroute(0);

static XrdOucTrace      DPMOfsTrace(&DPMOfsEroute);

/******************************************************************************/
/*                    F i l e   S y s t e m   O b j e c t                     */
/******************************************************************************/
         
#include "XrdDPM/XrdDPMOfs.icc"

/******************************************************************************/
/*                      L o c a l   V a r i a b l e s                         */
/******************************************************************************/

// some parameters determined at startup
//
static int foundparams = 0;
static char localhost[MAXHOSTNAMELEN];
static struct passwd stagersuperuser;
static char dpnshost[CA_MAXHOSTNAMELEN+1];
static int sessionEst = 0;
static time_t sessionTime;
static const int maxsessage = 20;

// this is the 'mad' prefix which is generally set by the oss
// layer when creating a N2N object. We make a local copy in a
// global for this module, so that the DPM OFS layer can use it
//
static char l_madpfx[512];
static int l_madpfxlen = 0;

// for mapping surl->sfn for currently open files
//
struct map_element_s {
   char sfn[CA_MAXSFNLEN+1];
   int count;
};
typedef std::map< std::string, struct map_element_s > sfnMap_t;
typedef sfnMap_t::iterator sfnMapItr;
static sfnMap_t sfnMap;
static XrdSysMutex sfnMapMutex;

// for queue of dpns and dpm operations
//
enum qtype { open_authz, getdone, putdone, dostat, abortfiles };
struct element_s {
   element_s *next;
   qtype type;
   const char *surl, *sfn, *r_token;
   XrdSecEntity* sec;
   int flags, retc;
   struct stat statbuf;
   XrdSysCondVar mutex;
   int done;
   element_s() : mutex(0) {}
};
static struct element_s *qroot = NULL;
static int numinQ = 0, isWaiting = 0;
static XrdSysCondVar queueMutex(0);

// identities
//
struct id_info_s {
   int nbfqans;
   uid_t uid;
   gid_t gids[MAXFQANS];
   char voname[CA_MAXCSECNAMELEN+1];
   char principal[CA_MAXCSECNAMELEN+1];
   char fqans[MAXFQANS][CA_MAXCSECNAMELEN+1];
   char *fqans_ptr[MAXFQANS];
} id_default, id_new, id_current;


/******************************************************************************/
/*                      L o c a l   F u n c t i o n s                         */
/******************************************************************************/
/******************************************************************************/
/*                      i s _ i d _ d i f f e r e n t                         */
/******************************************************************************/
static int is_id_different(const struct id_info_s *i1, const struct id_info_s *i2)
{
   if (strcmp(i1->principal, i2->principal)) return 1;
   if (strcmp(i1->voname, i2->voname)) return 1;
   if (i1->nbfqans != i2->nbfqans) return 1;
   for(int i=0;i<i1->nbfqans;++i)
     if (strcmp(i1->fqans[i],i2->fqans[i])) return 1;
   return 0;
}

/******************************************************************************/
/*                             c o n c a t _ f n                              */
/******************************************************************************/
static int concat_fn(const char *prefix, // String to prefix path
                     const int   pfxlen, // Length of prefix string  
                     const char *path,   // String to suffix prefix
                     char *buffer, // Resulting buffer
                     int   blen)   // The buffer length
{
   int addslash = (*path != '/');
   int pathlen  = strlen(path);

   if ((pfxlen + addslash + pathlen) >= blen) return -1;

   if (pfxlen) {strcpy(buffer, prefix); buffer += pfxlen;}
   if (addslash) {*buffer = '/'; buffer++;}
   strcpy(buffer, path);
   return 0;
}

/******************************************************************************/
/*                    d o _ g e t _ p a t h _ 2 _ s f n                       */
/******************************************************************************/
static int do_get_path_2_sfn(const char *path, char *out_sfn, size_t out_size) {
   std::string p(path);
   sfnMapMutex.Lock();
   sfnMapItr itr = sfnMap.find(p);
   if (itr != sfnMap.end()) {
      if (strlen(itr->second.sfn) >= out_size) {
         sfnMapMutex.UnLock();
         return -1;
      }
      strcpy(out_sfn, itr->second.sfn);
      sfnMapMutex.UnLock();
   } else {
      sfnMapMutex.UnLock();
      if (strlen(path) >= out_size)
         return -1;
      strcpy(out_sfn, path);
   }
   return 0;
}

/******************************************************************************/
/*                d o _ i n s e r t _ p a t h _ 2 _ s f n                     */
/******************************************************************************/
static void do_insert_path_2_sfn(const char *path, const char *sfn) {
   std::string p(path);
   sfnMapMutex.Lock();
   sfnMapItr itr = sfnMap.find(p);
   if (itr == sfnMap.end()) {
      struct map_element_s e;
      strncpy(e.sfn,sfn,CA_MAXSFNLEN); e.sfn[CA_MAXSFNLEN]='\0';
      e.count = 1;
      sfnMap[p] = e;
   } else {
      itr->second.count++;
   }
   sfnMapMutex.UnLock();
}

/******************************************************************************/
/*                d o _ r e m o v e _ p a t h _ 2 _ s f n                     */
/******************************************************************************/
static void do_remove_path_2_sfn(const char *path) {
   std::string p(path);
   sfnMapMutex.Lock();
   sfnMapItr itr = sfnMap.find(p);
   if (itr != sfnMap.end()) {
      if ( ! --itr->second.count ) {
         sfnMap.erase(itr);
      }
   }
   sfnMapMutex.UnLock();
}

/******************************************************************************/
/*                            d o _ d e q u e u e                             */
/******************************************************************************/
static int do_dequeue(element_s *req) {
   element_s *lp=0,*p;
   int ret = 0;
   queueMutex.Lock();
   p = qroot;
   while(p && !ret) {
      if (p == req) {
         if (lp) {
            lp->next = p->next;
         } else {
            qroot = p->next;
         }
         ret = 1;
         numinQ--;
      }
      lp = p;
      p = p->next;
   }
   queueMutex.UnLock();
   return ret;
}

/******************************************************************************/
/*             d o _ a d d _ t o _ q u e u e _ a n d _ w a i t                */
/******************************************************************************/
static int do_add_to_queue_and_wait(element_s *n) {
   // add to queue
   //
   queueMutex.Lock();
   element_s *e = qroot;
   n->next = 0;
   while(e && e->next) e=e->next;
   if (!e) {
      qroot = n;
   } else {
      e->next = n;
   }
   numinQ++;
   if (isWaiting) {
      isWaiting=0;
      queueMutex.Signal();
   }
   n->mutex.Lock();
   n->done = 0;
   queueMutex.UnLock();

   // wait for action to finish
   //
   int retc;
   do {
      retc = n->mutex.Wait(30);
      // continue while timeout but can not dequeue
      // (i.e. request is already being processed)
      // or signaled but not done flag not set (spurious wakeup)
   } while((retc && !do_dequeue(n)) || (!retc && !n->done));
   if (n->done) retc=n->retc;
   else retc = ETIME;

   n->mutex.UnLock();

   return retc;
}

/******************************************************************************/
/*                 d o _ r e m o v e _ f r o m _ q u e u e                    */
/******************************************************************************/
static element_s *do_remove_one_from_queue() {
   element_s *ret = 0;
   queueMutex.Lock();
   if (qroot) {
      ret = qroot;
      qroot = qroot->next;
      ret->next = 0;
      numinQ--;
   }
   queueMutex.UnLock();
   return ret;
}

/******************************************************************************/
/*                         c l o s e S e s s i o n                            */
/******************************************************************************/
static int closeSession() {
   if (sessionEst) {
      dpns_endsess();
      sessionEst = 0;
   }
   return 0;
}

/******************************************************************************/
/*                           a g e S e s s i o n                              */
/******************************************************************************/
static int ageSession(const time_t &t) {
   if (sessionEst && (sessionTime>t || t-sessionTime > maxsessage)) {
      dpns_endsess();
      sessionEst = false;
   }
   return 0;
} 

static int ageSession() {
   time_t t = time(NULL);
   return ageSession(t);
}

/******************************************************************************/
/*                 s e t u p A n d U s e S e s s i o n                        */
/******************************************************************************/
static int setupAndUseSession() {
   time_t t = time(NULL);

   if (is_id_different(&id_new,&id_current)) {
      closeSession();
      if (id_new.uid) {
         memcpy(&id_current, &id_new, sizeof(id_new));
         for(int i=0;i < id_current.nbfqans;i++)
            id_current.fqans_ptr[i] = id_current.fqans[i];
         dpns_client_setAuthorizationId (id_current.uid, id_current.gids[0], "GSI", id_current.principal);
         if (id_current.nbfqans)
            dpns_client_setVOMS_data(id_current.voname, id_current.fqans_ptr, id_current.nbfqans);
         dpm_client_setAuthorizationId (id_current.uid, id_current.gids[0], "GSI", id_current.principal);
         if (id_current.nbfqans)
            dpm_client_setVOMS_data(id_current.voname, id_current.fqans_ptr, id_current.nbfqans);
      } else {
        // can not reset dpm authorization!
        Cns_client_resetAuthorizationId();
      }
   } else
      ageSession(t);

   if (!sessionEst) {
      std::ostringstream o;
      std::string s;
      o << XrdDPMOfsCVSID;
#if defined(XRDDPM_GCC_VERSION)
      o << ": GCC version " << XRDDPM_GCC_VERSION;
#endif
      o << ": XrdDPM version " << XrdDPMVERSIONSTR;
#if defined (XrdCompileVERSION)
      o << ": Xrootd version " << XrdCompileVERSION;
#endif
#if defined (XRDDPM_BUILDDATE)
      o << ": Built " << XRDDPM_BUILDDATE;
#endif
      s = o.str();
      if (dpns_startsess(dpnshost, (char *)s.c_str())<0)
         return serrno;
      sessionEst = 1;
   }

   if (sessionEst)
      sessionTime = t;

   return 0;
}

/******************************************************************************/
/*                            d o _ g e t d o n e                             */
/******************************************************************************/
static int do_getdone(const char *surl, const char *token) {
   const char *tident = "";
   static const char *epname = "do_getdone";
   char *sp,s[CA_MAXPATHLEN+1],r[CA_MAXDPMTOKENLEN+1];
   struct dpm_filestatus *filestatuses = NULL;
   int retc, err, nbreplies = 0;

   if (strlen(surl)>CA_MAXPATHLEN) return ENAMETOOLONG;
   if (strlen(token)>CA_MAXDPMTOKENLEN) return EINVAL;

   strcpy(s,surl);
   strcpy(r,token);
   sp = s;

   retc = dpm_relfiles(r, 1, &sp, 0, &nbreplies, &filestatuses);
   err = serrno;

   dpm_free_filest(nbreplies, filestatuses);

   ZTRACE(close, "release files surl=" << surl
      << " request token " << token <<
      " return code " << retc << " error code " << err);

   if (retc<0) {
      return err;
   }

   return 0;
}

/******************************************************************************/
/*                            d o _ p u t d o n e                             */
/******************************************************************************/
static int do_putdone(const char *surl, const char *token) {
   const char *tident = "";
   static const char *epname = "do_putdone";
   char *sp,s[CA_MAXPATHLEN+1],r[CA_MAXDPMTOKENLEN+1];
   struct dpm_filestatus *filestatuses = NULL;
   int retc, err, nbreplies = 0;

   if (strlen(surl)>CA_MAXPATHLEN) return ENAMETOOLONG;
   if (strlen(token)>CA_MAXDPMTOKENLEN) return EINVAL;

   strcpy(s,surl);
   strcpy(r,token);
   sp = s;

   retc=dpm_putdone(r, 1, &sp, &nbreplies, &filestatuses);
   err = serrno;

   dpm_free_filest(nbreplies, filestatuses);

   ZTRACE(close, "put done surl=" << surl <<
      " request token " << token <<
      " return code " << retc << " error code " << err);

   if (retc<0) {
      return err;
   }

   return 0;
}

/******************************************************************************/
/*                        d o _ a b o r t f i l e s                           */
/******************************************************************************/
static int do_abortfiles(const char *surl, const char *token) {
   const char *tident = "";
   static const char *epname = "do_abortfiles";
   char *sp,s[CA_MAXPATHLEN+1],r[CA_MAXDPMTOKENLEN+1];
   struct dpm_filestatus *filestatuses = NULL;
   int retc, err, nbreplies = 0;

   if (strlen(surl)>CA_MAXPATHLEN) return ENAMETOOLONG;
   if (strlen(token)>CA_MAXDPMTOKENLEN) return EINVAL;

   strcpy(s,surl);
   strcpy(r,token);
   sp = s;

   retc=dpm_abortfiles(r, 1, &sp, &nbreplies, &filestatuses);
   err = serrno;

   dpm_free_filest(nbreplies, filestatuses);

   ZTRACE(close, "abort files surl=" << surl <<
      " request token " << token <<
      " return code " << retc << " error code " << err);

   if (retc<0) {
      return err;
   }

   return 0;
}

/******************************************************************************/
/*                 d o _ f i n d _ l o c a l p a r a m s                      */
/******************************************************************************/
static int do_find_localparams() {
   char localdomain[MAXHOSTNAMELEN];
   if (foundparams) return 0;
   if (gethostname(localhost,sizeof(localhost))) return -1;
   if (Cdomainname(localdomain,sizeof(localdomain))<0) return -1;
   if (!strchr(localhost, (int)'.')) {
      if (strlen(localhost)+strlen(localdomain)+2 > sizeof(localhost)) return -1;
      strcat(localhost,".");
      strcat(localhost, localdomain);
   }

   struct passwd *pw;
   if (*XrdDPMOfsFS.m_stagersuperuser) pw = getpwnam(XrdDPMOfsFS.m_stagersuperuser);
   else pw = getpwuid(geteuid());

   if (!pw) return -1;
   memcpy(&stagersuperuser, pw, sizeof(struct passwd));

   dpnshost[0] = '\0';
   char *val = getenv("DPNS_HOST");
   if (val && strlen(val)<=CA_MAXHOSTNAMELEN)
      strcpy(dpnshost, val);

   foundparams=1;
   return 0;
}

/******************************************************************************/
/*              d o _ d e c o d e _ e s c a p e d _ c g i                     */
/******************************************************************************/
static int do_decode_escaped_cgi(char *out, size_t outs, const char *p, size_t ins)
{
   if (!out || !outs || !p) return EINVAL;

   char *q = out;
   size_t iremaining = ins;
   size_t oremaining = outs;

   while(*p && iremaining && oremaining) {
      if (*p != '%') { *q++ = *p++; iremaining--; oremaining--; }
      else {
         if (iremaining < 3) return EINVAL;
         char c,buf[5],*r;
         strcpy(buf,"0x"); strncat(buf, p+1, 2); buf[4] = '\0';
         c = (char)strtol(buf, &r, 16);
         if (!r || *r) return EINVAL;
         *q++ = c;
         iremaining -= 3;
         p += 3;
         oremaining--;
      }
   }
   if (!oremaining || iremaining) return EINVAL;
   *q++ = '\0';

   return 0;
}

/******************************************************************************/
/*                    d o _ e x t r a c t _ d p m _ c g i                     */
/******************************************************************************/
static int do_extract_dpm_cgi(const char *ipath, const char *iinfo,
                              char *opath, size_t opath_size,
                              char *oinfo, size_t oinfo_size,
                              char *thost, size_t thost_size,
                              bool &using_cgi, char *sfn,
                              char *r_token, bool &is_put)
{
   if (!ipath || !opath || !oinfo ||
       !opath_size || !oinfo_size || !sfn || !r_token) return EINVAL;

   const char *p,*p2;

   // look for first '?' or '&dpm.' to see if there is any cgi data in ipath
   p = strstr(ipath, "?");
   p2 = strstr(ipath, "&dpm.");
   if (p && p2) p = (p<p2) ? p : p2;
   else p = (p) ? p : p2;

   p2 = &ipath[strlen(ipath)];
   oinfo[0] = '\0';

   if (p) {
      // copy first (i.e. path) part of name to opath
      //
      size_t plen = p - ipath;
      if (plen+1 > opath_size) return EINVAL;
      strncpy(opath, ipath, plen); opath[plen] = '\0';

      // copy cgi part of name to oinfo
      //
      p++;
      plen = p2-p;
      if (plen+2 > oinfo_size) return EINVAL;
      // make first char of oinfo a '&' then tag on the cgi
      oinfo[0]='&';
      strncpy(oinfo+1, p, plen); oinfo[plen+1] = '\0';
   } else {
      // no cgi in ipath so copy it all to opath
      //
      if (strlen(ipath)+1>opath_size) return EINVAL;
      strcpy(opath, ipath);
   }

   // add any explicit info to oinfo
   //
   if (iinfo && *iinfo) {
      if (strlen(oinfo)+strlen(iinfo)+2 > oinfo_size) return EINVAL;
      strcat(oinfo, "&");
      strcat(oinfo, iinfo);
   }

   // remove '?' or duplicated '&' characters from the info string
   //
   {
      char *r;
      while((r=strstr(oinfo,"?"))) {
         if (r[1] && r[1] != '&') r[0]='&';
         else memmove(r, r+1, strlen(r));
      }
      // replace duplicate & with a single &
      r=oinfo;
      while(*r++) {
        char *r2 = r;
        while(*r2=='&') r2++;
        if (*(r-1)=='&' && r2!=r) memmove(r,r2,1+strlen(r2));
      }
   }

   // assume no DPM cgi information
   //
   using_cgi = false;

   // look for DPM specific cgi info
   //
   p = strstr(oinfo, "&dpm.sfn=");
   p2 = strstr(oinfo, "&dpm.tok=");

   if (p && p2) {
      size_t s;
      const char *p3;
      // pick out escaped sfn from cgi info
      //
      p += 9; p3 = p;
      while(*p3 && *p3 != '&' && *p3 != '?') p3++;
      s = p3-p;
      if (do_decode_escaped_cgi(sfn, CA_MAXSFNLEN+1, p, s)) return EINVAL;

      // pick out put/get indication from token portion of cgi info
      //
      p2 += 9;
      if (*p2 != 'g' && *p2 != 'p') return EINVAL;
      is_put = (*p2=='p') ? true : false;
     
      // pick out the token from the cgi info
      //
      p2++; p3 = p2;
      while(*p3 && *p3 != '&' && *p3 != '?') p3++;
      s = p3-p2;
      if (s > CA_MAXDPMTOKENLEN) return EINVAL;
      strncpy(r_token, p2, s);
      r_token[s] = '\0';

      using_cgi = true;
   }

   if (thost && thost_size) {
     thost[0] = '\0';
   }

   if ((p=strstr(sfn, ":/"))) {
      size_t s = p-sfn;
      if (thost) {
         if (s>=thost_size) {
            return ENAMETOOLONG;
         }
         strncpy(thost, sfn, s);
         thost[s] = '\0';
      }
      memmove(sfn, p+1, strlen(sfn)-s);
   }

   return 0;
}

/******************************************************************************/
/*                             d o _ s t a t                                  */
/******************************************************************************/
static int do_stat(const char *path, struct stat *st)
{
   const char *tident = "";
   static const char *epname = "do_stat";
   const char *sfn = path;
   struct dpns_filestatg statbuf;
   char sfn1[CA_MAXSFNLEN+1];
   int retc;

   if (stat(sfn,st)<0) {
      int e = errno;
      ZTRACE(stat, "failed to perform stat() on " << sfn << " error code " << e);
      serrno = e;
      return -1;
   }

   if ((st->st_mode & S_IFMT) != S_IFREG ||
       (st->st_uid != stagersuperuser.pw_uid && st->st_gid != stagersuperuser.pw_gid)) {
      // not a file maager by DPM
      //
      ZTRACE(stat, "not a regular file or not"
         " owned by or in the group of the stagersuperuser(" << stagersuperuser.pw_uid <<
         "," << stagersuperuser.pw_gid << ") for file " << sfn);
      serrno = EPERM;
      return -1;
   }

   if (strlen(localhost)+strlen(sfn)+1 > CA_MAXSFNLEN) {
      ZTRACE(stat, "host:name too long for statr " << localhost << ":" << sfn);
      serrno = ENAMETOOLONG;
      return -1;
   }
   strcpy(sfn1, localhost);
   strcat(sfn1, ":");
   strcat(sfn1, sfn);

   if ((retc=setupAndUseSession())) {
      ZTRACE(stat, "could not setup DPNS session prior to dpns_statr for " << sfn1 << " return code " << retc);
      serrno = retc;
      return -1;
   }

   if (dpns_statr(sfn1, &statbuf) <0) {
      int e = serrno;
      ZTRACE(stat, "dpns_statr failed for " << sfn1 << " with error code " << e);
      serrno = e;
      return -1;
   }

   st->st_uid = statbuf.uid;
   st->st_gid = statbuf.gid;
   st->st_mode = statbuf.filemode;

   ZTRACE(stat, "dpns_statr succeeded for " << sfn1);

   return 0;
}

/******************************************************************************/
/*                   d o _ o p e n _ p e r m c h e c k                        */
/******************************************************************************/
static int do_open_permcheck(const char *path, const int &flags)
{
   const char *tident = "";
   static const char *epname = "do_open_permcheck";
   struct stat st;
   const char *sfn = path;
   char *parent,*p;
   char sfn1[CA_MAXSFNLEN+1];
   int retc;

   // expects globals: stagesuperuser localhost localdomain
   //
   if (stat (sfn, &st) < 0) {
      if (errno != ENOENT || (flags & O_CREAT) == 0) {
         int e = errno;
         ZTRACE(open, "error during permissions check (stat) for open of "
            << sfn << " error code " << e);
         serrno = e;
         return (-1);
      }
      if ((parent = strdup (sfn)) == NULL) {
         serrno = errno;
         return (-1);
      }
      if ((p = strrchr (parent, '/')) == NULL) {
         free (parent);
         serrno = EINVAL;
         return (-1);
      }
      if (p != parent) {
         *p = '\0';      
         p = parent;
      } else
         p = NULL;
      if (stat (p ? p : "/", &st) < 0) {
         int e = errno;
         ZTRACE(open,"error during permissions check (stat on reduced path)"
            " for open of " << sfn << " error code " << e);
         serrno = e;
         free (parent);
         return (-1);
      }
      free (parent);
   }
   if (st.st_uid != stagersuperuser.pw_uid && st.st_gid != stagersuperuser.pw_gid) {
      // not a file managed by DPM
      //
      ZTRACE(open,"during permissions check for open of " << sfn
         << " appears not to be owned or in group of stagersuperuser("
         << stagersuperuser.pw_uid << "," << stagersuperuser.pw_gid << ")");
      serrno = EPERM;
      return(-1); // don't allow
   }

   if (strlen(localhost)+strlen(sfn)+1 > CA_MAXSFNLEN) {
      ZTRACE(open,"host:name too long during preparation for open access check of "
         << localhost << ":" << sfn);
      serrno = ENAMETOOLONG;
      return -1;
   }
   strcpy(sfn1, localhost);
   strcat(sfn1, ":");
   strcat(sfn1, sfn);

   if ((retc=setupAndUseSession())) {
      ZTRACE(open,"Could not set session in preparation for dpns_accessr "
         "check during open of " << sfn1 << " error code " << retc);
      serrno = retc;
      return -1;
   }

   if (dpns_accessr (sfn1, (flags & (O_WRONLY|O_TRUNC|O_CREAT)) ? W_OK : R_OK) < 0) {
      int e = serrno;
      ZTRACE(open,"Access check for " << sfn1 << " failed, error code " << e);
      serrno = e;
      return (-1);
   }

   ZTRACE(open,"access check for " << sfn1 << " OK");
   return(0);
}

/******************************************************************************/
/*                 x r o o t d _ l o a d F i l e s y s t e m                  */
/******************************************************************************/       
static XrdSfsFileSystem *XrdXrootdloadFileSystem(XrdSysError *eDest,
                                                 char *fslib, const char *cfn)
{
   void *libhandle;
   XrdSfsFileSystem *(*ep)(XrdSfsFileSystem *, XrdSysLogger *, const char *);
   XrdSfsFileSystem *FS;
       
// Open the file system library
// 
   if (!(libhandle = dlopen(fslib, RTLD_NOW)))
      {eDest->Emsg("Config",dlerror(),"opening shared library",fslib);
       return 0;
      }

// Get the file system object creator
//
   if (!(ep = (XrdSfsFileSystem *(*)(XrdSfsFileSystem *,XrdSysLogger *,const char *))
                                  dlsym(libhandle,"XrdSfsGetFileSystem")))
      {eDest->Emsg("Config", dlerror(),
                   "finding XrdSfsGetFileSystem() in", fslib);
       return 0;
      }
 
// Get the file system object
//
   if (!(FS = (*ep)(0, eDest->logger(), cfn)))
      {eDest->Emsg("Config", "Unable to create file system object via",fslib);
       return 0;
      }

// All done
//
   return FS;
}

/******************************************************************************/
/*                         X r d D P M O f s C o m m                          */
/******************************************************************************/
static void *XrdDPMOfsComm(void *noarg)
{
   int retc,used_sec;
   element_s *p;
   static const char *epname = "XrdDPMOfsComm";
   const char *tident = "";

   dpm_seterrbuf(XrdDPMOfsFS.m_apibuf, sizeof(XrdDPMOfsFS.m_apibuf));
   dpns_seterrbuf(XrdDPMOfsFS.m_apibuf, sizeof(XrdDPMOfsFS.m_apibuf));

   uid_t effective_uid = geteuid();
   if ((effective_uid == 0 && strcmp(XrdDPMOfsFS.m_stagersuperuser,"root"))
        || effective_uid != stagersuperuser.pw_uid) {
      DPMOfsEroute.Emsg("XrdDPMOfsComm","Running as root without explicit stagersuperuser or mismatch with specified user");
      _exit(8);
   }

   memset(&id_default, '\0', sizeof(id_default));
   memset(&id_current, '\0', sizeof(id_current));
   memset(&id_new, '\0', sizeof(id_new));

   strcpy(id_default.principal, XrdDPMOfsFS.m_principal);

   if (*XrdDPMOfsFS.m_principal && strcmp(XrdDPMOfsFS.m_principal,"root") != 0) {
      id_default.nbfqans = (XrdDPMOfsFS.m_nbfqans > MAXFQANS) ? MAXFQANS : XrdDPMOfsFS.m_nbfqans;
      strcpy(id_default.voname, XrdDPMOfsFS.m_voname);
      if (!*XrdDPMOfsFS.m_voname) id_default.nbfqans = 0;
      const char *fqans[MAXFQANS];
      for(int i=0;i<id_default.nbfqans;i++) {strcpy(id_default.fqans[i],XrdDPMOfsFS.m_fqans[i]);fqans[i]=id_default.fqans[i];}
      if (dpns_getidmap (XrdDPMOfsFS.m_principal, id_default.nbfqans, fqans, &id_default.uid, id_default.gids)) {
         const char *err = sstrerror(serrno);
         const char *prefix = "Could not map principal to DPM internal ID, ";
         char *myerr = (char*)malloc(strlen(err)+strlen(prefix)+1);
         strcpy(myerr,prefix);
         strcat(myerr,err);
         DPMOfsEroute.Emsg("XrdDPMOfsComm", myerr);
         free(myerr);
         _exit(8);
      }
   }

   memcpy(&id_new, &id_default, sizeof(id_new));
   used_sec = 0;

   while(1) {
      queueMutex.Lock();
      if (!numinQ) {
         isWaiting = 1;
         retc = queueMutex.Wait(30);
         isWaiting = 0;
         queueMutex.UnLock();
      } else {
         queueMutex.UnLock();
      }

      p = do_remove_one_from_queue();
      //      TRACES("Took one from queue ");

      // some restrictions: To allow connections without XrdSec information we require that
      // a default principal was defined. (root is ok). However once an XrdSec connection has
      // been made, a non XrdSec conneciton can only be accepted later if the default
      // identity is not the super user. (Because of limitations in resetting the dpm authorizationId)
      //
      if (p) {
        if (!p->sec) {
          if (!*id_default.principal || (used_sec && !id_default.uid)) {
             p->mutex.Lock();
             p->retc = EPERM;
             p->done = 1;
             p->mutex.Signal();
             p->mutex.UnLock();
             ageSession();
             continue;
          }
        }
      }
      
      if (p && p->sec) {
	// we need to map if the server is configured to provide security information, e.g. sec entity is != 0 
        //
        struct id_info_s nid;
	int ret=0;
        memset(&nid, '\0', sizeof(nid));
        if (p->sec->name)
          strncpy(nid.principal, p->sec->name, sizeof(nid.principal)-1);
        if (p->sec->vorg)
           strncpy(nid.voname, p->sec->vorg, sizeof(nid.voname)-1);
        if (p->sec->role)
           strncpy(nid.fqans[0], p->sec->role,sizeof(nid.fqans[0])-1);
	if (*nid.fqans[0] && *nid.voname)
	  nid.nbfqans = 1;
        if (is_id_different(&id_current, &nid)) {
          if (is_id_different(&id_current, &id_default)) {
             closeSession();
             memcpy(&id_new, &id_default, sizeof(id_new));
          }
          setupAndUseSession();
          const char *fqans[1];
          fqans[0] = nid.fqans[0];
          if ((ret = dpns_getidmap (nid.principal, nid.nbfqans, fqans, &nid.uid, nid.gids))) {
	     ZTRACE(open, "error during mapping of "
		 << nid.principal << " to fqan " << nid.fqans[0] << " error code " << ret);
             serrno = ret;
             p->mutex.Lock();
             p->retc = serrno;
             p->done = 1;
             p->mutex.Signal();
             p->mutex.UnLock();
             ageSession();
             continue;
          }
          closeSession();
          memcpy(&id_new, &nid, sizeof(id_new));
          setupAndUseSession();
          used_sec = 1;
        }
      }	else if (p) {
         if (is_id_different(&id_current, &id_default)) {
            closeSession();
            memcpy(&id_new, &id_default, sizeof(id_new));
            setupAndUseSession();
         }
      }

      if (p) {
	if (p->type == open_authz) {
	  //	  TRACES("Doing open_authz");
	  retc = do_open_permcheck(p->sfn, p->flags);
	  if (retc<0) retc=serrno;
	}
	if (p->type == getdone) {
	  //	  TRACES("Doing getdone");
	  retc = do_getdone(p->surl, p->r_token);
	}
	if (p->type == putdone) {
	  //	  TRACES("Doing putdone");
	  retc = do_putdone(p->surl, p->r_token);
	  }
	if (p->type == dostat) {
	  //	  TRACES("Doing dostat");
            retc = do_stat(p->sfn, &p->statbuf);
            if (retc<0) retc=serrno;
	}
	if (p->type == abortfiles) {
		//	  TRACES("Doing dpm_abortfiles");
		retc = do_abortfiles(p->surl, p->r_token);
	}
	//	TRACES("Returning "<<retc);
        p->mutex.Lock();
        p->retc = retc;
        p->done = 1;
        p->mutex.Signal();
        p->mutex.UnLock();
      }

      ageSession();

   }
   return (void *)0;
}

/******************************************************************************/
/*                      X r d O u c N a m e 2 N a m e                         */
/******************************************************************************/
extern "C"
{
XrdOucName2Name *XrdOucgetName2Name(XrdOucgetName2NameArgs)
{
   XrdDPMN2N *inst = new XrdDPMN2N(eDest, lroot, rroot);

  if (parms) eDest->Say("++++++ XrdDPMN2N initializing. Local lfn prefix '",parms,"'.");
   else
     eDest->Say("++++++ XrdDPMN2N initializing. Local lfn prefix is null");

   if (inst->set_mad(parms)) {
      eDest->Say("++++++ XrdDPMN2N failed while setting prefix");
      delete inst;
      inst = NULL;
   }

   return (XrdOucName2Name *)inst;
}
}

/******************************************************************************/
/*                  X r d S f s G e t F i l e S y s t e m                     */
/******************************************************************************/
extern "C"
{
XrdSfsFileSystem *XrdSfsGetFileSystem(XrdSfsFileSystem *native_fs, 
                                      XrdSysLogger     *lp,
                                      const char       *configfn)
{
   pthread_t tid;
   int retc;

// Do the herald thing
//
   DPMOfsEroute.SetPrefix("XrdDPMOfs_");
   DPMOfsEroute.logger(lp);
   DPMOfsEroute.Emsg("Init", "(c) 2006 Stanford University/SLAC and CERN, XrdDPMOfs Version "
                          XrdDPMVERSIONSTR);

// Record the configuration file in our own instance
//
   XrdDPMOfsFS.ConfigFN = (configfn && *configfn ? strdup(configfn) : 0);

// configure: the null secondard lib pointer prevents configure
// attempting to configure the secondary
//
   XrdDPMOfsFS.m_nextXrdOfsFS = (XrdSfsFileSystem*)0;
   if ( XrdDPMOfsFS.Configure(DPMOfsEroute) ) return 0;

// This will execute the 'XrdSfsGetFileSystem' of m_FSLib
//
   if (*XrdDPMOfsFS.m_FSLib) {
      DPMOfsEroute.Emsg("Init", "Loading and initalising secondary OFS library");
      XrdDPMOfsFS.m_nextXrdOfsFS = XrdXrootdloadFileSystem(&DPMOfsEroute, XrdDPMOfsFS.m_FSLib, configfn);
   } else {
      XrdDPMOfsFS.m_nextXrdOfsFS = (XrdSfsFileSystem*)0;
   }

// Find some parameters needed for the dpns/dpm communication
   if (do_find_localparams()) {
      DPMOfsEroute.Emsg("XrdDPMOfsinit",EINVAL,"Unable to determine local hostname or stager superuser identity.");
      return 0;
   }

   DPMOfsEroute.Emsg("XrdOfsinit","final initialisation the XrdDPMOfs");

// Start a thread to make the authz, getdone, putdone calls to dpns/dpm
//
   if ((retc = XrdSysThread::Run(&tid, XrdDPMOfsComm, (void *)0)))
      DPMOfsEroute.Emsg("XrdDPMOfsinit", retc, "create DPM communication thread");

   return (XrdSfsFileSystem*) &XrdDPMOfsFS;
}
}

/******************************************************************************/
/*                                                                            */
/*           D i r e c t o r y   O b j e c t   I n t e r f a c e s            */
/*                                                                            */
/******************************************************************************/
/******************************************************************************/
/*                        C o n s t r u c t o r                               */
/******************************************************************************/
XrdDPMOfsDirectory::XrdDPMOfsDirectory(XrdSfsFileSystem *xf, char* user)
   : XrdSfsDirectory(user)
{
   if (xf)
      m_nextXrdOfsDirectory = xf->newDir(user);
   else
      m_nextXrdOfsDirectory = (XrdSfsDirectory *)0;

   tident = user;
}

/******************************************************************************/
/*                          D e s t r u c t o r                               */
/******************************************************************************/
XrdDPMOfsDirectory::~XrdDPMOfsDirectory()
{
   if (m_nextXrdOfsDirectory) delete m_nextXrdOfsDirectory;
}

/******************************************************************************/
/*                                 o p e n                                    */
/******************************************************************************/
int XrdDPMOfsDirectory::open(const char              *dirName,
                             const XrdSecEntity      *client,
                             const char              *opaque)
{
   static const char *epname = "opendir";
   return XrdDPMOfsFS.Emsg(epname, error, EOPNOTSUPP, "open directory");
}

/******************************************************************************/
/*                            n e x t E n t r y                               */
/******************************************************************************/
const char *XrdDPMOfsDirectory::nextEntry()
{
   return NULL;
}

/******************************************************************************/
/*                               c l o s e                                    */
/******************************************************************************/
int XrdDPMOfsDirectory::close()
{
   static const char *epname = "closedir";
   return XrdDPMOfsFS.Emsg(epname, error, EOPNOTSUPP, "close directory");
}

/******************************************************************************/
/*                               F N a m e                                    */
/******************************************************************************/
const char *XrdDPMOfsDirectory::FName()
{
   return NULL;
}

/******************************************************************************/
/*                                                                            */
/*                F i l e   O b j e c t   I n t e r f a c e s                 */
/*                                                                            */
/******************************************************************************/
/******************************************************************************/
/*                        C o n s t r u c t o r                               */
/******************************************************************************/
XrdDPMOfsFile::XrdDPMOfsFile(XrdSfsFileSystem *xf, char* user)
   : XrdSfsFile(user)
{
   m_inUse = false;
   m_info = NULL;

   if (xf)
      m_nextXrdOfsFile = xf->newFile(user);
   else
      m_nextXrdOfsFile = (XrdSfsFile *)0;

   tident = user;
}

/******************************************************************************/
/*                          D e s t r u c t o r                               */
/******************************************************************************/
XrdDPMOfsFile::~XrdDPMOfsFile()
{
   if (m_nextXrdOfsFile) delete m_nextXrdOfsFile;
   m_nextXrdOfsFile=0;
   clearInUse(true);
   if (m_info) free(m_info);
}

/******************************************************************************/
/*                                  o p e n                                   */
/******************************************************************************/
int XrdDPMOfsFile::open(const char          *path,      // In
                        XrdSfsFileOpenMode   open_mode, // In
                        mode_t               Mode,      // In
                  const XrdSecEntity        *client,    // In
                  const char                *opaque)    // In
{
   static const char *epname = "open";
   bool usingCgi, isPut;
   char rToken[CA_MAXDPMTOKENLEN+1];
   char url[CA_MAXPATHLEN+1],sfn[CA_MAXSFNLEN+1];
   char *newinfo;
   const char *epath;
   const char *info = opaque;
   int flags, retc;
   size_t s;

   if (!m_nextXrdOfsFile) {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "open file");
   }

// Split off any cgi information and see if there is DPM cgi info
//
   s = ((info)?strlen(info):0)+strlen(path)+1;
   newinfo = (char*)malloc(s);
   if (do_extract_dpm_cgi(path, info, url, sizeof(url),
         newinfo, s, NULL, 0, usingCgi, sfn, rToken, isPut)) {
      free(newinfo);
      return XrdDPMOfsFS.Emsg(epname, error, EINVAL, "invalid turl", path);
   }

   if (usingCgi) {
      ZTRACE(open,"open surl=" << url << " with cgi information "
         << "sfn=" << sfn << " " << (isPut ? "put" : "get") << " token=" << rToken
         << " opaque data=" << newinfo);
   } else {
      ZTRACE(open,"open turl=" << url << " opaque data=" << newinfo);
   }

// don't need to use
//
   free(newinfo);

// Set the actual open mode
//
   XrdSfsFileOpenMode om = open_mode;
   if (open_mode & SFS_O_CREAT) om = SFS_O_CREAT;
      else if (open_mode & SFS_O_TRUNC) om = SFS_O_TRUNC;

   flags = (open_mode & SFS_O_RESET ? O_EXCL : 0);
   switch(om & (SFS_O_RDONLY | SFS_O_WRONLY | SFS_O_RDWR |
                       SFS_O_CREAT  | SFS_O_TRUNC))
   {
      case SFS_O_RDONLY:
         flags = O_RDONLY;
         break;

      case SFS_O_WRONLY:
         flags = O_WRONLY;
         break;

      case SFS_O_RDWR:
         flags = O_RDWR;
         break;

      case SFS_O_CREAT:
         flags = O_RDWR|O_CREAT|O_TRUNC|O_EXCL;
         break;

      case SFS_O_TRUNC:
         flags |= O_RDWR|O_CREAT|O_TRUNC;
         break;

      default:
         flags = O_RDONLY;
         break;
   }

   // check permissions
   //
   retc = 0;
   if (usingCgi) {
      if (!isPut && flags != O_RDONLY) {
         ZTRACE(open,"presenting a get token with flags other than read only");
         retc = EPERM;
      }
   }

// epath should be the physical storage name
// either the sfn from the cgi information
// or otherwise a tidied version of the path
//
   epath = (usingCgi) ? sfn : url;

   //   TRACES("Opening Path " << epath);
   if (!retc) {
      element_s *p = new element_s;
      p->type = open_authz;
      p->sfn = epath;
      p->flags = flags;
      if (XrdDPMOfsFS.m_useclientid)
         p->sec = (XrdSecEntity*)client;
      else
         p->sec = (XrdSecEntity*)0;
      retc = do_add_to_queue_and_wait(p);
      delete p;
   }

   // insert a mapping for the namemapper:
   //
   do_insert_path_2_sfn(url, epath);
      
   if (retc) {
      retc = XrdDPMOfsFS.Emsg(epname, error, retc, "to access", path);
   } else {
      m_nextXrdOfsFile->error = error;
      // pass open the 'tidied' path, i.e. with any &dpm suffixes removed
      // as open may do an authorization check based on the path. Pass
      // the original info.
      retc = m_nextXrdOfsFile->open(url, open_mode, 0660, client, info);
      error = m_nextXrdOfsFile->error;
   }

   if (retc == SFS_OK) {
      clearInUse(false);
      m_client = client;
      strncpy(m_fullpath, path, CA_MAXPATHLEN); m_fullpath[CA_MAXPATHLEN]='\0';
      strncpy(m_tidypath, url, CA_MAXPATHLEN); m_tidypath[CA_MAXPATHLEN]='\0';
      strncpy(m_sfn, epath, CA_MAXSFNLEN); m_sfn[CA_MAXSFNLEN]='\0';
      m_usingCgi = usingCgi;
      if (usingCgi) {
         strcpy(m_rToken, rToken);
         m_isPut = isPut;
      }
      if (m_info) free(m_info);
      m_info = 0;
      if (info)
         m_info = strdup(info);
      setInUse();
   } else {
      // do not release files here; the open may have failed due to authorization error
      // in which case we can not trust the request token
      // doDPMDone(usingCgi, isPut, url, rToken, true);
   }

// remove the entra count in the mapper
//
   do_remove_path_2_sfn(url);

   return retc;
}

/******************************************************************************/
/*                                 c l o s e                                  */
/******************************************************************************/
int XrdDPMOfsFile::close()
{
   static const char *epname = "close";
   if (!m_nextXrdOfsFile) {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "close file");
   }

   m_nextXrdOfsFile->error = error;
   int retc = m_nextXrdOfsFile->close();
   error = m_nextXrdOfsFile->error;

   clearInUse(false);

   return retc;
}

/******************************************************************************/
/*                               F N a m e                                    */
/******************************************************************************/
const char *XrdDPMOfsFile::FName()
{
   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      const char *retc = m_nextXrdOfsFile->FName();
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else
      return NULL;
}

/******************************************************************************/
/*                             g e t M m a p                                  */
/******************************************************************************/
int XrdDPMOfsFile::getMmap(void **Addr, off_t &Size)
{
   static const char *epname = "getMmap";
//
// Set the size to zero so that even if there is an error
// the size will stil indicate there is no mapping
//
   Size = 0;

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->getMmap(Addr, Size);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "get memory mapped file");
   }
}

/******************************************************************************/
/*                                 r e a d                                    */
/******************************************************************************/
int XrdDPMOfsFile::read(XrdSfsFileOffset   fileOffset,
                        XrdSfsXferSize     preread_sz)
{
   static const char *epname = "read";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->read(fileOffset,preread_sz);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "read file");
   }
}

XrdSfsXferSize XrdDPMOfsFile::read(XrdSfsFileOffset   fileOffset,
                                   char              *buffer,
                                   XrdSfsXferSize     buffer_size)
{
   static const char *epname = "read";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      XrdSfsXferSize retc = m_nextXrdOfsFile->read(fileOffset,buffer,buffer_size);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "read file");
      return -1;
   }
}

int XrdDPMOfsFile::read(XrdSfsAio *aioparm)
{
   static const char *epname = "read";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->read(aioparm);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "read file");
   }
}

/******************************************************************************/
/*                                w r i t e                                   */
/******************************************************************************/
XrdSfsXferSize XrdDPMOfsFile::write(XrdSfsFileOffset   fileOffset,
                                    const char        *buffer,
                                    XrdSfsXferSize     buffer_size)
{
   static const char *epname = "write";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      XrdSfsXferSize retc = m_nextXrdOfsFile->write(fileOffset,buffer,buffer_size);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "write file");
      return -1;
   }
}

int XrdDPMOfsFile::write(XrdSfsAio *aioparm)
{
   static const char *epname = "write";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->write(aioparm);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "write file");
   }
}

/******************************************************************************/
/*                                  s y n c                                   */
/******************************************************************************/
int XrdDPMOfsFile::sync()
{
   static const char *epname = "sync";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->sync();
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "sync file");
   }
}
        
int XrdDPMOfsFile::sync(XrdSfsAio *aiop)
{
   static const char *epname = "sync";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->sync(aiop);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "sync file");
   }
}

/******************************************************************************/
/*                                  s t a t                                   */
/******************************************************************************/
int XrdDPMOfsFile::stat(struct stat *buf)
{
   static const char *epname = "stat";
   
   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->stat(buf);
      error = m_nextXrdOfsFile->error;
      if (retc) return retc;

      // make a stat call to the dpns
      //
      element_s *p = new element_s;
      p->type = dostat;
      p->sfn = m_sfn;
      p->sec = 0;
      retc = do_add_to_queue_and_wait(p);

      if (!retc) {
         memcpy(buf,&p->statbuf,sizeof(struct stat));
      }
      delete p;
      return retc;
   }
   else {
      return XrdDPMOfsFS.stat(m_fullpath, buf, error, m_client, m_info);
   }

}

/******************************************************************************/
/*                              t r u n c a t e                               */
/******************************************************************************/
int XrdDPMOfsFile::truncate(XrdSfsFileOffset   fileOffset)
{
   static const char *epname = "truncate";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->truncate(fileOffset);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "truncate file");
   }
}

/******************************************************************************/
/*                            g e t C X i n f o                               */
/******************************************************************************/
int XrdDPMOfsFile::getCXinfo(char cxtype[4], int &cxrsz)
{
   static const char *epname = "getCXinfo";

   if (m_nextXrdOfsFile) {
      m_nextXrdOfsFile->error = error;
      int retc = m_nextXrdOfsFile->getCXinfo(cxtype,cxrsz);
      error = m_nextXrdOfsFile->error;
      return retc;
   }
   else {
      return XrdDPMOfsFS.Emsg(epname, error, ENOTCONN, "get file CX info");
   }
}

/******************************************************************************/
/*                            s e t I n U s e                                 */
/******************************************************************************/
void XrdDPMOfsFile::setInUse()
{
   do_insert_path_2_sfn(m_tidypath, m_sfn);
   m_inUse = true;
}

/******************************************************************************/
/*                             d o D P M D o n e                              */
/******************************************************************************/
void XrdDPMOfsFile::doDPMDone(const bool &isError) const
{
   doDPMDone(m_usingCgi, m_isPut, m_tidypath, m_rToken, isError);
}

void XrdDPMOfsFile::doDPMDone(const bool &usingCgi, const bool &isPut,
                              const char *path, const char *token,
                              const bool &isError) const
{
   if (!usingCgi) return;

   char buff1[1024],*buff;
   int blen;
   element_s *p = new element_s;

   if (!isPut)
     p->type = getdone;
   else if (!isError)
    p->type = putdone;
   else
    p->type = abortfiles;
   p->r_token = token;
   p->sec = 0;
   buff = buff1; blen = sizeof(buff1);
   if ( l_madpfxlen && strstr(path, l_madpfx) != path ) {
      if (concat_fn(l_madpfx, l_madpfxlen, path, buff, blen)) {
         blen = strlen(path)+l_madpfxlen+2;
         buff = (char*)malloc(blen);
         if (concat_fn(l_madpfx, l_madpfxlen, path, buff, blen)) {
            delete p;
            free(buff);
            return;
         }
      }
   } else {
      if (strlen(path)>=blen) {
         blen = strlen(path)+1;
         buff = (char*)malloc(blen);
      }
      strcpy(buff, path);
   }
   p->surl = buff;
   (void)do_add_to_queue_and_wait(p);
   delete p;
   if (buff != buff1) free(buff);
}

/******************************************************************************/
/*                           c l e a r I n U s e                              */
/******************************************************************************/
void XrdDPMOfsFile::clearInUse(const bool &isError)
{
   if (!m_inUse) return;
   do_remove_path_2_sfn(m_tidypath);
   doDPMDone(isError);
   m_inUse = false;
}

/******************************************************************************/
/*                                                                            */
/*         F i l e   S y s t e m   O b j e c t   I n t e r f a c e s          */
/*                                                                            */
/******************************************************************************/
/******************************************************************************/
/*                                c h m o d                                   */
/******************************************************************************/
int XrdDPMOfs::chmod(const char             *Name,
                           XrdSfsMode        Mode,
                           XrdOucErrInfo    &out_error,
                     const XrdSecEntity     *client,
                     const char             *opaque)
{
   static const char *epname = "chmod";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "change access mode");
}

/******************************************************************************/
/*                              e x i s t s                                   */
/******************************************************************************/
int XrdDPMOfs::exists(const char                *fileName,
                            XrdSfsFileExistence &exists_flag,
                            XrdOucErrInfo       &out_error,  
                      const XrdSecEntity        *client,
                      const char                *opaque)
{
   static const char *epname = "exists";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "existence check");
}

/******************************************************************************/
/*                                f s c t l                                   */
/******************************************************************************/
int XrdDPMOfs::fsctl(const int               cmd,
                     const char             *args,
                           XrdOucErrInfo    &out_error,
                     const XrdSecEntity     *client)
{
   static const char *epname = "fsctl";
   int opcode = cmd & SFS_FSCTL_CMD;
   if (opcode == SFS_FSCTL_LOCATE)
      {
       char locResp[2048];
       char rType[3], *Resp[] = {rType, locResp};
       rType[0] = 'S';
       rType[1] = 'r';
       rType[2] = '\0';
       sprintf(locResp,"[::%s] ",dpnshost);
       out_error.setErrInfo(strlen(locResp)+3, (const char **)Resp, 2);
       return SFS_DATA;
      }
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "fsctl", args);
}

/******************************************************************************/
/*                           g e t S t a t s                                  */
/******************************************************************************/
int XrdDPMOfs::getStats(char *buff, int blen)
{
   return 0;
}

/******************************************************************************/
/*                         g e t V e r s i o n                                */
/******************************************************************************/
const char *XrdDPMOfs::getVersion()
{
   if (m_nextXrdOfsFS)
      return m_nextXrdOfsFS->getVersion();
   else {
      return NULL;
   }
}

/******************************************************************************/
/*                                 m k d i r                                  */
/******************************************************************************/
int XrdDPMOfs::mkdir(const char             *dirName,
                           XrdSfsMode        Mode,   
                           XrdOucErrInfo    &out_error,
                     const XrdSecEntity     *client,
                     const char             *opaque)
{
   static const char *epname = "mkdir";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "make directory");
}

/******************************************************************************/
/*                              p r e p a r e                                 */
/******************************************************************************/
int XrdDPMOfs::prepare(XrdSfsPrep       &pargs,
                       XrdOucErrInfo    &out_error,
                 const XrdSecEntity     *client)
{
   static const char *epname = "prepare";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "prepare");
}

/******************************************************************************/
/*                                   r e m                                    */
/******************************************************************************/
int XrdDPMOfs::rem(const char             *path,
                         XrdOucErrInfo    &out_error,
                   const XrdSecEntity     *client,
                   const char             *info)
{
   static const char *epname = "rem";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "rem");
}

/******************************************************************************/
/*                              r e m d i r                                   */
/******************************************************************************/
int XrdDPMOfs::remdir(const char             *dirName,
                            XrdOucErrInfo    &out_error,
                      const XrdSecEntity     *client,
                      const char             *info)
{
   static const char *epname = "remdir";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "remdir");
}

/******************************************************************************/
/*                              r e n a m e                                   */
/******************************************************************************/
int XrdDPMOfs::rename(const char             *oldFileName,
                      const char             *newFileName,
                            XrdOucErrInfo    &out_error,  
                      const XrdSecEntity     *client,
                      const char             *infoO,
                      const char             *infoN)
{
   static const char *epname = "rename";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "rename");
}

/******************************************************************************/
/*                                  s t a t                                   */
/******************************************************************************/
int XrdDPMOfs::stat(const char             *path,        // In   
                          struct stat      *buf,         // Out
                          XrdOucErrInfo    &einfo,       // Out
                    const XrdSecEntity     *client,      // In
                    const char             *info)        // In
{
   const char *tident = einfo.getErrUser();
   static const char *epname = "stat";
   char rToken[CA_MAXDPMTOKENLEN+1];
   char url[CA_MAXPATHLEN+1],sfn[CA_MAXSFNLEN+1];
   bool usingCgi, isPut;
   char *newinfo;
   const char *epath; 
   int retc;
   size_t s;

   if (!m_nextXrdOfsFS) {
      return XrdDPMOfsFS.Emsg(epname, einfo, ENOTCONN, "stat");
   }
   
// Split off any cgi information and see if there is DPM cgi info
//
   s = ((info)?strlen(info):0)+strlen(path)+1;
   newinfo = (char*)malloc(s);
   if (do_extract_dpm_cgi(path, info, url, sizeof(url),
         newinfo, s, NULL, 0, usingCgi, sfn, rToken, isPut)) {
      free(newinfo);
      return XrdDPMOfsFS.Emsg(epname, einfo, EINVAL, "invalid turl", path);
   }

// don't need to use
//
   free(newinfo);

   epath = (usingCgi) ? sfn : url;

 // insert a mapping for the namemapper:
 //
   do_insert_path_2_sfn(url, epath);

// Allow secondary OFS plugin to give an error:
// perhaps it applies some authorization checks. If the OSS associated
// with the secondary OFS has an error stating the real file we would also
// fail.
//
   retc = m_nextXrdOfsFS->stat(url, buf, einfo, client, info);
   if (retc != SFS_OK) {
      do_remove_path_2_sfn(url);
      ZTRACE(stat, "secondary OFS instance returned error on stat of " << path << " (" << url << ")");
      return retc;
   }

   do_remove_path_2_sfn(url);

// make a stat call to the dpns
//
   element_s *p = new element_s;
   p->type = dostat;
   p->sfn = epath;
   p->sec = (XrdSecEntity*)client;

   retc = do_add_to_queue_and_wait(p);

   if (!retc) {
      memcpy(buf,&p->statbuf,sizeof(struct stat));
   }
   delete p;

   if (retc) {
      return XrdDPMOfsFS.Emsg(epname, einfo, retc, "stat");
   }

   return SFS_OK;
}

int XrdDPMOfs::stat(const char             *path,        // In   
                          mode_t           &mode,        // Out
                          XrdOucErrInfo    &einfo,       // Out
                    const XrdSecEntity     *client,      // In
                    const char             *info)        // In
{
   struct stat buf;
   int retc = stat(path, &buf, einfo, client, info);

   if (retc == SFS_OK) {
      mode = buf.st_mode;
   }

   return retc;
}

/******************************************************************************/
/*                            t r u n c a t e                                 */
/******************************************************************************/
int XrdDPMOfs::truncate(const char*, XrdSfsFileOffset,
			XrdOucErrInfo& out_error,
			const XrdSecEntity* entity,
			const char* info) 
{
   static const char *epname = "truncate";
   return XrdDPMOfsFS.Emsg(epname, out_error, EOPNOTSUPP, "truncate");
}

/******************************************************************************/
/*                                x t r a c e                                 */
/******************************************************************************/
int XrdDPMOfs::xtrace(XrdOucStream &Config, XrdSysError &Eroute)
{
    static struct traceopts {const char *opname; int opval;} tropts[] =
       {
        {"aio",      TRACE_aio},
        {"all",      TRACE_ALL},
        {"chmod",    TRACE_chmod},
        {"close",    TRACE_close},
        {"closedir", TRACE_closedir},
        {"debug",    TRACE_debug},
        {"delay",    TRACE_delay},
        {"dir",      TRACE_dir},
        {"exists",   TRACE_exists},  
        {"getstats", TRACE_getstats},
        {"fsctl",    TRACE_fsctl},
        {"io",       TRACE_IO},
        {"mkdir",    TRACE_mkdir},
        {"most",     TRACE_MOST},
        {"open",     TRACE_open},
        {"opendir",  TRACE_opendir},
        {"qscan",    TRACE_qscan},
        {"read",     TRACE_read},
        {"readdir",  TRACE_readdir},
        {"redirect", TRACE_redirect},
        {"remove",   TRACE_remove},
        {"rename",   TRACE_rename},
        {"sync",     TRACE_sync},
        {"truncate", TRACE_truncate},
        {"write",    TRACE_write}
       };
    int i, neg, trval = 0, numopts = sizeof(tropts)/sizeof(struct traceopts);
    char *val;
                   
    if (!(val = Config.GetWord()))
       {Eroute.Emsg("Config", "trace option not specified"); return 1;}
    while (val)
         {if (!strcmp(val, "off")) trval = 0;
             else {if ((neg = (val[0] == '-' && val[1]))) val++;
                   for (i = 0; i < numopts; i++)
                       {if (!strcmp(val, tropts[i].opname))
                           {if (neg) trval &= ~tropts[i].opval;
                               else  trval |=  tropts[i].opval;
                            break;   
                           }
                       }
                   if (i >= numopts)
                      Eroute.Emsg("Config", "invalid trace option -", val);
                  }
          val = Config.GetWord(); 
         }
    DPMOfsTrace.What = trval;
         
// All done
//
   return 0;
}

/******************************************************************************/
/*                           C o n f i g u r e                                */
/******************************************************************************/
int XrdDPMOfs::Configure(XrdSysError &Eroute)
{
   if (m_nextXrdOfsFS) {
      return 0;
   }

   m_principal[0] = '\0';
   m_nbfqans = 0;
   m_voname[0] = '\0';
   m_FSLib[0] = '\0';
   m_stagersuperuser[0] = '\0';
   m_useclientid = 1;

   umask(0);

   const char *var, *val;
   int  cfgFD, retc, NoGo = 0;
   XrdOucEnv myEnv;
   XrdOucStream Config(&Eroute, getenv("XRDINSTANCE"), &myEnv, "=====> ");
   int v;

   if( !ConfigFN || !*ConfigFN)
     Eroute.Emsg("Config", "Configuration file not specified.");
  else {
     // Try to open the configuration file.
     //
     if ( (cfgFD = open(ConfigFN, O_RDONLY, 0)) < 0)
        return Eroute.Emsg("Config", errno, "open config file", ConfigFN);
     Config.Attach(cfgFD);

     // Now start reading records until eof.
     //
     while((var = Config.GetMyFirstWord())) {
        if (!strncmp(var, "dpm.", 4)) {
           var += 4;

           if (!strcmp("surlprefix",var)) {
              if (!(val = Config.GetWord()) || strlen(val)>=sizeof(l_madpfx)) {
                 Eroute.Emsg("Config", "argument for surlprefix invalid.");NoGo |= 1;}
               else { 
                  if (l_madpfxlen && strcmp(val, l_madpfx)) {
                     Eroute.Emsg("Config", "surl prefix already set to a different value.");NoGo |= 1;}
                  else {strcpy(l_madpfx,val);l_madpfxlen=strlen(val);}
               }
           }
           if (!strcmp("fslib",var)) {
              if (!(val = Config.GetWord()) || strlen(val)>=sizeof(m_FSLib)) {
                 Eroute.Emsg("Config", "argument for fslib invalid.");NoGo |= 1;}
               else { strcpy(m_FSLib,val); }
           }
           if (!strcmp("useclientid",var)) {
              if (!(val = Config.GetWord())) {
                 Eroute.Emsg("Config", "argument for useclientid invalid.");NoGo |= 1;}
               else { if(XrdOuca2x::a2tm(Eroute, "useclientid value", val, &v))
                      NoGo |= 1; else m_useclientid = v;}
           }
           if (!strcmp("stagersuperuser",var)) {
              if (!(val = Config.GetWord()) || strlen(val)>=sizeof(m_stagersuperuser)) {
                 Eroute.Emsg("Config", "argument for stagersuperuser invalid.");NoGo |= 1;}
               else { strcpy(m_stagersuperuser,val); }
           }
        }
        if (!strncmp(var, "ofs.",4)) {
           var +=4;
           if (!strcmp("trace",var)) {
              NoGo |= xtrace(Config, Eroute);
           }
        }
     }
  
     // Now check if any errors occured during file i/o
     //
     if ((retc = Config.LastError()))
        NoGo = Eroute.Emsg("Config", -retc, "read config file", ConfigFN);
     Config.Close();

     // Now a dedicated scan for the authorization principal
     //
     if ( (cfgFD = open(ConfigFN, O_RDONLY, 0)) < 0)
        return Eroute.Emsg("Config", errno, "open config file", ConfigFN);
     Config.Attach(cfgFD);
     while((var = Config.GetLine())) {
        while(*var && *var==' ') var++;
        if (*var=='#') continue;
        if (!strncmp("dpm.principal ",var,14)) {
           var += 14;
           while(*var && *var==' ') var++;
           const char *tmp = var;
           while(*tmp) tmp++;
           while(tmp!=var && *(tmp-1)==' ') tmp--;
           size_t s = tmp-var;
           if (s>CA_MAXCSECNAMELEN) {Eroute.Emsg("Config","invalid argument for principal name."); NoGo |= 1;}
           else {strncpy(m_principal,var,s);m_principal[s]='\0';}
        }
        if (!strncmp("dpm.voname ",var,11)) {
           var += 11;
           while(*var && *var==' ') var++;
           const char *tmp = var;
           while(*tmp) tmp++;
           while(tmp!=var && *(tmp-1)==' ') tmp--;
           size_t s = tmp-var;
           if (s>CA_MAXCSECNAMELEN) {Eroute.Emsg("Config","invalid argument for voname name."); NoGo |= 1;}
           else {strncpy(m_voname,var,s);m_voname[s]='\0';}
        }
        if (!strncmp("dpm.fqan ",var,9)) {
           var += 9;
           while(*var && *var==' ') var++;
           const char *tmp = var;
           while(*tmp) tmp++;
           while(tmp!=var && *(tmp-1)==' ') tmp--;
           size_t s = tmp-var;
           if (m_nbfqans >= sizeof(m_fqans)/(CA_MAXCSECNAMELEN+1)) {Eroute.Emsg("Config","too many fqans."); NoGo |= 1;}
           else if (s>CA_MAXCSECNAMELEN) {Eroute.Emsg("Config","invalid argument for fqan name."); NoGo |= 1;}
           else {strncpy(m_fqans[m_nbfqans],var,s);m_fqans[m_nbfqans][s]='\0';m_nbfqans++;}
        }
     }
      if ((retc = Config.LastError()))
         NoGo = Eroute.Emsg("Config", -retc, "read config file", ConfigFN);
      Config.Close();
   }

   return NoGo;
}

/******************************************************************************/
/*                                  E m s g                                   */
/******************************************************************************/

int XrdDPMOfs::Emsg(const char    *pfx,    // Message prefix value
                 XrdOucErrInfo &einfo,  // Place to put text & error code
                 int            ecode,  // The error code
                 const char    *op,     // Operation being performed
                 const char    *target) // The target (e.g., fname)
{
   char *etext, buffer[XrdOucEI::Max_Error_Len], unkbuff[64];

// If the error is EBUSY then we just need to stall the client. This is
// a hack in order to provide for proxy support
//
    if (ecode < 0) ecode = -ecode;
    if (ecode == EBUSY) return 5;  // A hack for proxy support

// Get the reason for the error
//
   if (!(etext = DPMOfsEroute.ec2text(ecode)))
      {sprintf(unkbuff, "reason unknown (%d)", ecode); etext = unkbuff;}

// Format the error message
//
    snprintf(buffer,sizeof(buffer),"Unable to %s %s; %s", op, target, etext);   
   
// Print it out if debugging is enabled
//
#ifndef NODEBUG
    DPMOfsEroute.Emsg(pfx, einfo.getErrUser(), buffer);
#endif
  
// Place the error message in the error object and return
//
    einfo.setErrInfo(ecode, buffer);
    return SFS_ERROR;
}

/******************************************************************************/
/*                           C o n s t r u c t o r                            */
/******************************************************************************/

XrdDPMN2N::XrdDPMN2N(XrdSysError *erp, const char *lpfx, const char *rpfx)
{  
   eDest = erp;
}

/******************************************************************************/
/*                             l f n 2 p f n                                  */
/******************************************************************************/
int XrdDPMN2N::lfn2pfn(const char *lfn, char *buff, int blen)
{
   if (do_get_path_2_sfn(lfn, buff, blen)) {
      return eDest->Emsg("XrdDPMN2N::lfn2pfn",-ENAMETOOLONG,"generate local path lfn=",lfn);
   }
   return 0;
}

/******************************************************************************/
/*                             l f n 2 r f n                                  */
/******************************************************************************/
int XrdDPMN2N::lfn2rfn(const char *lfn, char *buff, int blen)
{
   return eDest->Emsg("XrdDPMN2N::lfn2rfn",-EOPNOTSUPP,"generate remote path lfn=",lfn);
}

/******************************************************************************/
/*                             p f n 2 l f n                                  */
/******************************************************************************/
int XrdDPMN2N::pfn2lfn(const char *pfn, char *buff, int blen)
{
   return eDest->Emsg("XrdDPMN2N::pfn2lfn",-EOPNOTSUPP,"generate logical path pfn=",pfn);
}

/******************************************************************************/
/*                              s e t _ m a d                                 */
/******************************************************************************/

int XrdDPMN2N::set_mad(const char *mad)
{
   if (l_madpfxlen && (!mad || (*mad && strcmp(mad, l_madpfx)))) {
      eDest->Say("Attempting to set different local lfn prefix to that which is already set.");
      return 1;
   }

   // keep copy in global for use by DPM OFS
   //
   if (mad && *mad && strlen(mad)<sizeof(l_madpfx)) {
     strcpy(l_madpfx, mad);
     l_madpfxlen = strlen(mad);
   }
   return 0;
}
