/******************************************************************************/
/*                                                                            */
/*                          X r d D P M X m i . c c                           */
/*                                                                            */
/* DPM handling and various by David Smith, CERN                              */            
/* Other parts:                                                               */
/*                                                                            */
/* (c) 2007 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

static const char XrdDPMXmiCVSID[] = "$Id: XrdDPMXmi.cc,v 1.6 2010/01/14 16:08:13 dhsmith Exp $";

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "dpns_api.h"
#include "dpm_api.h"
#include "serrno.h"

#include <memory>
#include <iostream>
#include <sstream>
#include <string>

#include <dlfcn.h>
#ifndef __macos__
#include <link.h>
#endif

#include "XrdDPM/XrdDPMXmiQueueAndReq.hh"
#include "XrdDPM/XrdDPMXmi.hh"
#include "XrdDPM/XrdDPMXmiTrace.hh"
#include "XrdDPM/XrdDPMVersion.hh"

#include "Xrd/XrdScheduler.hh"
#include "XrdSys/XrdSysError.hh"
#include "XrdOuc/XrdOucName2Name.hh"
#include "XrdSys/XrdSysTimer.hh"
#include "XrdSys/XrdSysPlatform.hh"
#include "XrdOuc/XrdOucStream.hh"
#include "XrdOuc/XrdOuca2x.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdCms/XrdCmsTypes.hh"
#include "XrdSfs/XrdSfsInterface.hh"

#define DPMRETRY(x) ((x)==SECOMERR || (x)==SEINTERNAL)
#define XPROTO "xroot"
#define ESCAPEURLCHAR(c) (c=='%' || c==':' || c=='?' || c=='&' || c==' ')

/******************************************************************************/
/*                               G l o b a l s                                */
/******************************************************************************/

XrdSysError  *XrdDPMXmi::eDest;
XrdInet      *XrdDPMXmi::iNet;
XrdScheduler *XrdDPMXmi::Sched ;
XrdOucTrace  *XrdDPMXmi::Trace;
XrdOucName2Name *XrdDPMXmi::N2N;
static char l_madpfx[512];
static int l_madpfxlen = 0;

// couple of local functions
static int defaultLfn2rfn(const char *Path, char *new_path, size_t s);
static char *prefix_mad(const char *surl, char *out, size_t out_len);

/******************************************************************************/
/*                         L o c a l   C l a s s e s                          */
/******************************************************************************/

namespace DPMXmi
{
// This class is used to re-initialize the external initerface. It is always
// runasynchnrously by another thread is is scheduled whenever initialization
// is needed (e.g., after a communications failure).
//
class XmiInit : XrdJob
{
public:

void DoIt() {XmiP->InitXeq(); delete this;}

     XmiInit(XrdDPMXmi *xp) : XrdJob("DPM xmi init") {XmiP = xp;}
    ~XmiInit() {}

private:

XrdDPMXmi *XmiP;
};

// This class is used to provide an error string which contains function name
// and the concatenation of the error from the dpns/dpm API plus, an optional
// alternate fixed buffer and a supplied error string
//
// Note: construction and print() truncates the api & alternate buffers
//
class XmiErr
{
public:

void clear() {
        if (apibuf) apibuf[0]='\0';
        if (altbuf) altbuf[0]='\0';
     }

char*print(const char *fserror) {
        size_t len = sizeof(totstring);
        size_t remain = len - 1;

        strcpy(totstring, fname); remain -= strlen(fname);
        if (apibuf) {
           while(*apibuf && apibuf[strlen(apibuf)-1]=='\n') apibuf[strlen(apibuf)-1]='\0';
           if (strlen(apibuf)>0 && remain>3) {
             strcat(totstring, ": "); remain -= 2;
           }
           strncat(totstring,apibuf,remain);
           totstring[sizeof(totstring)-1] = '\0';
           remain = len - strlen(totstring) - 1 ;
        }
        if (altbuf) {
           while(*altbuf && altbuf[strlen(altbuf)-1]=='\n') altbuf[strlen(altbuf)-1]='\0';
           if (strlen(altbuf)>0 && remain>3) {
             strcat(totstring, ": "); remain -= 2;
           }
           strncat(totstring,altbuf,remain);
           totstring[sizeof(totstring)-1] = '\0';
           remain = len - strlen(totstring) - 1 ;
        }
        if (fserror) {
           if (strlen(fserror)>0 && remain>3) {
             strcat(totstring, ": "); remain -= 2;
           }
           strncat(totstring,fserror,remain);
           totstring[sizeof(totstring)-1] = '\0';
        }
        clear();
        return totstring;
     }

     XmiErr(const char *name, char *apibuf, char *altbuf) {
        fname[0] = '\0';
        if (name) {
           strncpy(fname, name, sizeof(fname));
           fname[sizeof(fname)-1] = '\0';
        }
        this->apibuf = apibuf;
        this->altbuf = altbuf;
        clear();
     }
    ~XmiErr() {}

private:
     char fname[256];
     char totstring[1024];
     char *apibuf,*altbuf;
};
}

/******************************************************************************/
/*                               G l o b a l s                                */
/******************************************************************************/

/******************************************************************************/
/*            E x t e r n a l   T h r e a d   I n t e r f a c e s             */
/******************************************************************************/
  
void *XrdDPMXmi_StartPoll(void *parg)
{  
   XrdDPMXmi *requestProcessor = (XrdDPMXmi *)parg;

   requestProcessor->MSSPoll();

   return (void *)0;
}

/******************************************************************************/
/*                          X r d O l b g e t X m i                           */
/******************************************************************************/

// We allow the the following on the xmi directive line after the library path:

// [host[:port]]

// Where: host:port is the location of the DPM name server

extern "C"
{
XrdCmsXmi *XrdCmsgetXmi(int argc, char **argv, XrdCmsXmiEnv *XmiEnv)
{

// Get an Xmi Object and return it
//
   return (XrdCmsXmi *)(new XrdDPMXmi(XmiEnv));
}
}
  
/******************************************************************************/
/*                           C o n s t r u c t o r                            */
/******************************************************************************/
  
XrdDPMXmi::XrdDPMXmi(XrdCmsXmiEnv *Env)
{
   char *tp, *Parms, *cp;

// Set pointers and variables to zero
//
   initDone     = 0;
   initActive   = 0;

// Copy out needed stuff from the environment to globals
//
   eDest = Env->eDest;         // -> Error message handler
   iNet  = Env->iNet;          // -> Network object
   Sched = Env->Sched;         // -> Thread scheduler
   Trace = Env->Trace;         // -> Trace handler
   N2N = Env->Name2Name;       // -> lfn to xxx mapper (may be null)

// Prefil the Opts structure with defaults and fill with any
// values which might be specified in the config file
//
   SetDefaultsAndReadConfig(Env);

// Load authorization lib if authopt is enabled
//
   if (Opts.authopt) {
     void *libhandle;
     XrdAccAuthorize *(*ep)(XrdSysLogger *, const char *, const char *);
     if (!(libhandle = dlopen(Opts.authlib,RTLD_NOW))) {
       eDest->Emsg("Config",dlerror(),"opening shared library",Opts.authlib);
       _exit(8);
     }
     ep = (XrdAccAuthorize *(*)(XrdSysLogger *, const char *, const char *))
                             (dlsym(libhandle,"XrdAccAuthorizeObject"));
     if (!ep) {
       eDest->Emsg("Config","Error finding XrdAccAuthorizeObject in",Opts.authlib);
       _exit(8);
     }
     if (!(Opts.authorization = ep(eDest->logger(), Env->ConfigFN, *Opts.authparm ? Opts.authparm : 0))) {
       eDest->Emsg("Config","Failed to get an authorizaiton object from the entry point");
       _exit(8);
     }
   }
 
// no dpns session established yet
   m_dpnsSessEst = false;

// Obtain the opts data which come from the command line
//
   if (Env->Parms && *(Env->Parms))
   {
      Parms = strdup(Env->Parms); tp = Parms;
      while(*tp && *tp == ' ') tp++;
      if (*tp && *tp != '-')
      {
         if (strlen(tp)<=CA_MAXHOSTNAMELEN)
         {
            strcpy(Opts.dpns_host,tp); tp = index(tp, ' ');
            if ((cp = index(Opts.dpns_host, ':')))
            {
                // nb atoi not thread safe
               *cp = '\0'; Opts.dpns_port = atoi(cp+1);
            }
         }
      }
      free(Parms);
   }

// fill in m_ServerAndPort (dpns):
// possibly from the command line or the defaults, or nothing
//
   m_ServerAndPort[0] = '\0';
   if (Opts.dpns_host && strlen(Opts.dpns_host)<CA_MAXHOSTNAMELEN)
      strcpy(m_ServerAndPort,Opts.dpns_host);

   if (*m_ServerAndPort && Opts.dpns_port) {
      size_t needed, avail = sizeof(m_ServerAndPort) - strlen(m_ServerAndPort);
      int ret = snprintf(&m_ServerAndPort[strlen(m_ServerAndPort)], avail, ":%hu", Opts.dpns_port);
      needed = ret>=0 ? ret : 0;
      if (ret<0 || needed>=avail) {
         m_ServerAndPort[0]='\0';
      }
   }

// Propogate the setting to our queue
//
   queue = new XrdDPMQueue();
   queue->Set(this,Env);

// Schedule initialization of this object
//
   Init(0,1);
};

/******************************************************************************/
/*                             D e s t r u c t o r                            */
/******************************************************************************/
  
XrdDPMXmi::~XrdDPMXmi()
{
   if (queue) delete queue;
}

/******************************************************************************/
/*                               I n i t X e q                                */
/******************************************************************************/
  
void XrdDPMXmi::InitXeq()
{
   EPNAME("InitXeq");
   static int PollerOK = 0;
   pthread_t tid;
   int       retc;

// First we must verify that we haven't been initialized yet. There may be
// several threads that have noticed that re-initialization is needed.
//
   initMutex.Lock();
   if (initDone)
      {DEBUG("DPM Init skipped; already initialized.");
       initMutex.UnLock(); 
       return;
      }

// Start the thread that handles polling
//
   if (!PollerOK)
      if ((retc = XrdSysThread::Run(&tid, XrdDPMXmi_StartPoll, (void *)this,
                                     XRDSYSTHREAD_BIND, "dpm request polling")))
         eDest->Emsg("DPMXmi", retc, "create polling thread");
         else PollerOK = 1;

// All done
//
   initDone   = 1;
   initActive = 0;
   initMutex.UnLock();
   eDest->Say("XrdDPMXmi: DPM interface initialized.");
}
 
/******************************************************************************/
/*                               M S S P o l l                                */
/******************************************************************************/
  
void XrdDPMXmi::MSSPoll()
{
   int nq = 0, remain = -1;
   char *fqans_ptr[MaxFqans];

// Set the DPM & DPNS APIs to use the buffer in this XMI instance. This setting
// is thread specific for the APIs, but only if Cglobals is initialised.
//
   dpm_seterrbuf(m_apibuf, sizeof(m_apibuf));
   dpns_seterrbuf(m_apibuf, sizeof(m_apibuf));

// Set the authorization ID, voname, fqans in the dpns and dpm API, if specified
//
   if (*Opts.principal && strcmp(Opts.principal,"root") != 0) {

      int i;
      char *mech = "GSI";
      uid_t uid;
      gid_t gids[MaxFqans];

      if (!*Opts.voname) Opts.nbfqans = 0;
      for(i=0;i < Opts.nbfqans;i++)
         fqans_ptr[i] = Opts.fqans[i];

      setupAndUseSession();
      if (dpns_getidmap (Opts.principal, Opts.nbfqans, (const char **)fqans_ptr, &uid, gids)) {
         const char *err = sstrerror(serrno);
         const char *prefix = "Could not map principal to DPM internal ID, ";
         char *myerr = (char*)malloc(strlen(err)+strlen(prefix)+1);
         strcpy(myerr,prefix);
         strcat(myerr,err);
         eDest->Emsg("XrdDPMXmi::MSSPoll", myerr);
         free(myerr);
         _exit(8);
      }
      closeSession();
      if (uid) {
        dpns_client_setAuthorizationId (uid, gids[0], mech, Opts.principal);
        if (Opts.nbfqans)
          dpns_client_setVOMS_data(Opts.voname, fqans_ptr, Opts.nbfqans);
        dpm_client_setAuthorizationId (uid, gids[0], mech, Opts.principal);
        if (Opts.nbfqans)
          dpm_client_setVOMS_data(Opts.voname, fqans_ptr, Opts.nbfqans);
      }
   }

   while(1) {
      if (remain<0)
         remain = MSSPollTime;

      // wait for either something new or a fraction of MSSPollTime to pass
      queue->Wait4Q(nq, remain);

      // make sure all entries are have been serviced within the last MSSPollTime;
      // actions which are expected to happen faster should be checked within
      // a shorter time
      while (!(remain = queue->MakeOnePoll(nq)));

      // check if the (dpns) session is growing unused, close it if it is
      ageSession();
   }
}

/******************************************************************************/
/*                    F u n c t i o n a l   M e t h o d s                     */
/******************************************************************************/

/******************************************************************************/
/*                                 C h m o d                                  */
/******************************************************************************/
  
int XrdDPMXmi::Chmod(XrdCmsReq *Request, mode_t mode, const char *path, const char *opaque)
{
   pData1 *pd = new pData1;
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_chmod, Request, path, (void *)pd);
   if (!rp) { delete pd; return 1; }
   pd->mode = mode;
   rp->Queue();
   return 1;
}

int XrdDPMXmi::Chmod_process(XrdDPMReq *rp, const bool &destroy)
{
   EPNAME("Chmod_process");
   std::auto_ptr<pData1> pd((pData1*)rp->getPd());
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Chmod",m_apibuf,0);

   if (destroy) {
      sendError(Request, ETIME, "Chmod timedout", rp->Path());
      return 1;
   }

   if (setupAndUseSession(rp, error)) return 1;

// Try to change the mode
//
   if ( dpns_chmod(rp->Path(),pd->mode) < 0 ) {
      if (DPMRETRY(serrno)) {
         sendWait(Request, serrno, error.print(0));
         return 1;
      }
      sendError(Request, serrno, error.print(0), rp->Path());
   } else {
// Tell client all went well
//
      TRACE(Debug, "Sending Reply_OK for chmod");
      Request->Reply_OK();
   }

// Can now remove this request
   return 1;    
}
  
/******************************************************************************/
/*                                 M k d i r                                  */
/******************************************************************************/
  
int XrdDPMXmi::Mkdir(XrdCmsReq *Request, mode_t mode, const char *path, const char *opaque)
{
   pData1 *pd = new pData1;
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_mkdir, Request, path, (void *)pd);
   if (!rp) { delete pd; return 1; }
   pd->mode = mode;
   rp->Queue();
   return 1;
}

int XrdDPMXmi::Mkdir_process(XrdDPMReq *rp,const bool &destroy)
{
   std::auto_ptr<pData1> pd((pData1*)rp->getPd());
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Mkdir",m_apibuf,0);

   if (destroy) {
      sendError(Request, ETIME, "Mkdir timedout", rp->Path());
      return 1;
   }

   if (setupAndUseSession(rp, error)) return 1;

// Try to make the directory
//
   if ( dpns_mkdir(rp->Path(),pd->mode) < 0 ) {
      if (DPMRETRY(serrno)) {
         sendWait(Request, serrno, error.print(0));
         return 1;
      }
      sendError(Request, serrno, error.print(0), rp->Path());
   } else {
// Tell client all went well
//
      Request->Reply_OK();
   }

// Can now remove this request
   return 1;
}
  
/******************************************************************************/
/*                                M k p a t h                                 */
/******************************************************************************/

int XrdDPMXmi::Mkpath(XrdCmsReq *Request, mode_t mode, const char *path, const char *opaque)
{
   pData1 *pd = new pData1;
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_mkpath, Request, path, (void *)pd);
   if (!rp) { delete pd; return 1; }
   pd->mode = mode;
   rp->Queue();
   return 1;
}

int XrdDPMXmi::Mkpath_process(XrdDPMReq *rp, const bool &destroy)
{
   std::auto_ptr<pData1> pd((pData1*)rp->getPd());
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Mkpath",m_apibuf,0);
   int retc;

   if (destroy) {
      sendError(Request, ETIME, "Mkpath timedout", rp->Path());
      return 1;
   }

   if (setupAndUseSession(rp, error)) return 1;

   retc = mkp(rp->Path(), pd->mode);

   if (retc>0) {
      if (DPMRETRY(retc)) {
         sendWait(Request, retc, error.print(0));
         return 1;
      }
      sendError(Request, retc, error.print(0), rp->Path());
      return 1;
   }

   Request->Reply_OK();
   return 1;
}

/******************************************************************************/
/*                                R e n a m e                                 */
/******************************************************************************/
  
int XrdDPMXmi::Rename(XrdCmsReq *Request, const char *opath, const char *oldopaque,
                                          const char *npath, const char *newopaque)
{
   EPNAME("Rename");
   DPMXmi::XmiErr error("Rename",m_apibuf,0);
   int rc;
   TRACE(Debug, "Rename, opath=" << opath << " npath=" << npath);

   if (strlen(opath)>CA_MAXPATHLEN || strlen(npath)>CA_MAXPATHLEN)
      return sendError(Request, ENAMETOOLONG, error.print(0), strlen(opath)>CA_MAXPATHLEN ? opath : npath);

   pData2 *pd = new pData2;
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_mv, Request, "", (void *)pd);
   if (!rp) { delete pd; return 1; }

   if (N2N) {
     rc = N2N->lfn2rfn(opath, pd->opath, sizeof(pd->opath));
     if (!rc) rc = N2N->lfn2rfn(npath, pd->npath, sizeof(pd->npath));
   } else {
     rc = defaultLfn2rfn(opath, pd->opath, sizeof(pd->opath));
     if (!rc) rc = defaultLfn2rfn(npath, pd->npath, sizeof(pd->npath));
   }

   if (rc) { delete pd; return 1; }

   rp->Queue();
   return 1;
}

int XrdDPMXmi::Rename_process(XrdDPMReq *rp,const bool &destroy)
{
   std::auto_ptr<pData2> pd((pData2*)rp->getPd());
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Rename",m_apibuf,0);

   if (destroy) {
      sendError(Request, ETIME, "Rename timedout", pd->opath);
      return 1;
   }

   if (setupAndUseSession(rp, error)) return 1;
   
// Try to rename
//
   if ( dpns_rename(pd->opath, pd->npath)<0 ) {
      if (DPMRETRY(serrno)) {
         sendWait(Request, serrno, error.print(0));
         return 1;
      }
      sendError(Request, serrno, error.print(0), pd->opath);
      return 1;
   }

// Tell client all went well
//
   Request->Reply_OK();
   return 1;
}
  
/******************************************************************************/
/*                                R e m d i r                                 */
/******************************************************************************/
  
int XrdDPMXmi::Remdir(XrdCmsReq *Request, const char *path, const char *opaque)
{
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_rmdir, Request, path, 0);
   if (!rp) { return 1; }
   rp->Queue();
   return 1;
}

int XrdDPMXmi::Remdir_process(XrdDPMReq *rp, const bool &destroy)
{
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Remdir",m_apibuf,0);

   if (destroy) {
      sendError(Request, ETIME, "Remdir timedout", rp->Path());
      return 1;
   }

   if (setupAndUseSession(rp, error)) return 1;
   
// Try to remove the directory
//
   if ( dpns_rmdir(rp->Path())<0 ) {
      if (DPMRETRY(serrno)) {
         sendWait(Request, serrno, error.print(0));
         return 1;
      }
      sendError(Request, serrno, error.print(0), rp->Path());
      return 1;
   }

// Tell client all went well
//
   Request->Reply_OK();
   return 1;
}
  
/******************************************************************************/
/*                                R e m o v e                                 */
/******************************************************************************/
  
int XrdDPMXmi::Remove(XrdCmsReq *Request, const char *path, const char *opaque)
{
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_rm, Request, path, 0);
   if (!rp) { return 1; }
   rp->Queue();
   return 1;
}

int XrdDPMXmi::Remove_process(XrdDPMReq *rp, const bool &destroy)
{
   XrdCmsReq *Request = rp->Request();
   char *pp;
   struct dpm_filestatus *filestatuses=NULL;
   DPMXmi::XmiErr error("Remove",m_apibuf,0);
   int nbreplies = 0;
   int retc,ernum;

   if (destroy) {
      sendError(Request, ETIME, "Remove timedout", rp->Path());
      return 1;
   }

   pp = strdup(rp->Path());
   retc = dpm_rm(1, &pp, &nbreplies, &filestatuses);
   if (retc<0) ernum = serrno;

   free(pp);
   dpm_free_filest(nbreplies, filestatuses);

   if (retc<0) {
      if (DPMRETRY(ernum)) {
         sendWait(Request, ernum, error.print(0));
         return 1;
      }
      sendError(Request, ernum, error.print(0), rp->Path());
      return 1;
   }

// Tell client all went well
//
   Request->Reply_OK();

   return 1;
}
  
/******************************************************************************/
/*                                S e l e c t                                 */
/******************************************************************************/

int XrdDPMXmi::Select(XrdCmsReq *Request, int opts, const char *path, const char *opaque)
{
   pData3 *pd = new pData3;
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_stage, Request, path, (void *)pd);
   if (!rp) { delete pd; return 1; }

   // check authorization with authlib
   //
   if (Opts.authorization) {
     Access_Operation ao = AOP_Read;
     if (opts & (XMI_NEW | XMI_TRUNC)) ao = AOP_Create;
     else if (opts & XMI_RW) ao = AOP_Update;
     XrdOucEnv Open_Env(opaque);
     if (!Opts.authorization->Access((XrdSecEntity *)0, path, ao, &Open_Env)) {
       const char *err = rcToStr(EACCES);
       Request->Reply_Error(err,"Authorization failed");
       delete pd;
       rp->Recycle();
       return 1;
     }
   }

   // queue up request
   //
   pd->opts = opts;
   pd->tried_mkpath = false;
   pd->first = true;
   rp->Queue();
   return 1;
}


int XrdDPMXmi::Select_process(XrdDPMReq *rp, const bool &destroy)
{
   std::auto_ptr<pData3> pd((pData3*)rp->getPd());
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Select",m_apibuf,pd->fstatbuf);

   if (destroy) {
      sendError(Request, ETIME, "Select timedout", rp->Path());
      return 1;
   }

   if ((pd->opts & XMI_RW) && !(pd->opts & XMI_TRUNC) && !(pd->opts & XMI_NEW)) {
      notSupported(Request, "open for update",rp->Path());
      return 1;
   }

   if (pd->first) {
      if (pd->opts & (XMI_NEW | XMI_RW | XMI_TRUNC)) {
         return Select_process_put_first(rp, pd, error);
      } else {
         return Select_process_get_first(rp, pd, error);
      }
      // not reached
   }

   // not pd->first:
   //   check status of request

   if (pd->get) {
         return Select_process_get(rp, pd, error);
   } else {
         return Select_process_put(rp, pd, error);
   }
   // not reched
}

/******************************************************************************/
/*                                  S t a t                                   */
/******************************************************************************/

int XrdDPMXmi::Stat(XrdCmsReq *Request, const char *path, const char *opaque)
{
   XrdDPMReq *rp = queue->Alloc(XrdCmsXmiReq::do_stat, Request, path, 0);
   if (!rp) { return 1; }
   rp->Queue();
   return 1;
}

int XrdDPMXmi::Stat_process(XrdDPMReq *rp, const bool &destroy)
{
   XrdCmsReq *Request = rp->Request();
   DPMXmi::XmiErr error("Stat",m_apibuf,0);
   struct dpns_filestat dpnsbuf;
   struct stat statbuf;

   if (destroy) {
      sendError(Request, ETIME, "Stat timedout", rp->Path());
      return 1;
   }

   if (setupAndUseSession(rp, error)) return 1;
   
// Get information about the file
//
   if ( dpns_stat(rp->Path(),&dpnsbuf)<0 ) {
      if (DPMRETRY(serrno)) {
         sendWait(Request, serrno, error.print(0));
         return 1;
      }
      sendError(Request, serrno, error.print(0), rp->Path());
      return 1;
   }

// Convert the dpns stat buffer to a normal Unix one
//
   (void)memset((void *)&statbuf, 0, sizeof(statbuf));
   statbuf.st_mode  = dpnsbuf.filemode;
   statbuf.st_nlink = dpnsbuf.nlink;
   statbuf.st_uid   = dpnsbuf.uid;
   statbuf.st_gid   = dpnsbuf.gid;
   statbuf.st_size  = static_cast<off_t>(dpnsbuf.filesize);
   statbuf.st_atime = dpnsbuf.atime;
   statbuf.st_ctime = dpnsbuf.ctime;
   statbuf.st_mtime = dpnsbuf.mtime;
   if (dpnsbuf.status == '-') statbuf.st_ino = 1;

// Now send the response (let the request object figure out the format)
//
   Request->Reply_OK(statbuf);
   return 1;
}

/******************************************************************************/
/*                 g e t E x t r a S u r l O p t i o n s                      */
/******************************************************************************/

int XrdDPMXmi::getExtraSurlOptions(const char *p, const char *key, size_t maxlen,
                                   const char **outp, size_t *outlen) const
{
   if (!p || !key || !maxlen || !outp || !outlen) return EINVAL;

   char *k=(char*)malloc(7+strlen(key));
   strcpy(k,"&dpm.");
   strcat(k,key);
   strcat(k,"=");
   const char *r = strstr(p, k);

   if (!r) { free(k); return ENOENT; }

   r += strlen(k);
   free(k);

   const char *r2 = r;
   while(*r2 && *r2!='&' && *r2!='?') r2++;

   *outlen = r2 -r;
   *outp = r;

   if (*outlen > maxlen) return ENAMETOOLONG;

   return 0;
}

int XrdDPMXmi::getExtraSurlOptions(const char *Path, XrdDPMReq *rp) const
{
   if (!Path) return EINVAL;
   const char *p = index(Path, (int)'?');
   const char *p2 = strstr(Path, "&dpm.");
   if (p && p2) p = (p<p2) ? p : p2;
   else p = (p) ? p : p2;

   if (p) {
      char *p3 = (char *)malloc(strlen(p)+1);
      strcpy(p3,p);
      *p3 = '&';
      p = p3;
      size_t s;

      if (!getExtraSurlOptions(p,"stoken",CA_MAXDPMTOKENLEN, &p2, &s)) {
         rp->sopts.have_stoken = true;
         strncpy(rp->sopts.stoken, p2, s); rp->sopts.stoken[s] = '\0';
      }
      if (!getExtraSurlOptions(p,"ftype", 1, &p2, &s)) {
         rp->sopts.have_ftype = true;
         rp->sopts.ftype = (s) ? *p2 : '\0';
      }
      if (!getExtraSurlOptions(p,"lifetime",10, &p2, &s)) {
         char l[11];
         strncpy(l,p2,s); l[s]='\0';
         rp->sopts.have_lifetime = true;
         rp->sopts.lifetime = strtol(l, NULL, 0);
      }
      if (!getExtraSurlOptions(p,"reqsize",20, &p2, &s)) {
         char l[21];
         strncpy(l,p2,s); l[s]='\0';
         rp->sopts.have_reqsize = true;
         rp->sopts.reqsize = strtoll(l, NULL, 0);
      }
      free(p3);
   }
   return 0;
}

/******************************************************************************/
/*                     d e f a u l t L f n 2 r f n                            */
/******************************************************************************/

static int defaultLfn2rfn(const char *Path, char *new_path, size_t s)
{
   if (!Path || !new_path) return EINVAL;

   if (strncmp(Path,"srm:",4) && strncmp(Path,"sfn:",4)) {
      // remove anything after a '?' or a '&dpm.'
      //
      const char *p = index(Path, (int)'?');
      const char *p2 = strstr(Path, "&dpm.");
      if (p && p2) p = (p<p2) ? p : p2;
      else p = (p) ? p : p2;
     
      size_t l = (p) ? (p-Path) : strlen(Path);
      if (l>=s) return ENAMETOOLONG;
      strncpy(new_path, Path, l);
      new_path[l] = '\0';
      return 0;
   }

// ignore "srm:" or "sfn:"
   const char *p3,*p2,*p = &Path[4];

// skip any '/'
   while(*p=='/') ++p;

// skip anything (i.e. hostname) that is not '/'
   p2 = p; while(*p2 && *p2!='/') ++p2;

   if (!*p2) return EINVAL;

// check the hostname is reasonably formed
//
   size_t l = p2-p;
   if (l > CA_MAXHOSTNAMELEN) return ENAMETOOLONG;
   char host[CA_MAXHOSTNAMELEN+1];
   strncpy(host,p,l);
   host[l]='\0';
   if (!*host) return ENOENT;

   p3 = p2;

// remove anything after a '?' or a '&dpm.'
//
   p = index(p3, (int)'?');
   p2 = strstr(p3, "&dpm.");
   if (p && p2) p = (p<p2) ? p : p2;
   else p = (p) ? p : p2;

   l = (p) ? (p-p3) : strlen(p3);
   if (l>=s) return ENAMETOOLONG;
   strncpy(new_path, p3, l);
   new_path[l] = '\0';

   return 0;
}

int XrdDPMXmi::defaultLfn2rfn(const char *Path, char *new_path, size_t s) {
   char buffer1[CA_MAXPATHLEN+1];
   int rc = ::defaultLfn2rfn(Path, buffer1, sizeof(buffer1));

   if (rc) return rc;
   
   if ( l_madpfxlen && strstr(buffer1, l_madpfx) != buffer1 ) {
     if (!::prefix_mad(buffer1, new_path, s))
        return ENAMETOOLONG;
   } else {
     if (strlen(buffer1)>=s)
        return ENAMETOOLONG;
     strcpy(new_path, buffer1);
   }

   return 0;
}

/******************************************************************************/
/*                       P r i v a t e   M e t h o d s                        */
/******************************************************************************/

/******************************************************************************/
/*                                  I n i t                                   */
/******************************************************************************/
  
void XrdDPMXmi::Init(int deltatime, int force)
{
  EPNAME("DPMInit");
  XrdJob *jp = (XrdJob *)new DPMXmi::XmiInit(this);

// Turn off the initdone flag unless we are already running
//
   initMutex.Lock();
   if (force) initActive = 0;
      else if (initActive)
              {DEBUG("DPM Init not scheduled; already started.");
               initMutex.UnLock();
               return;
              }
   initDone = 0;
   initMutex.UnLock();

// Schedule this initialization job as needed
//
   if (deltatime) Sched->Schedule(jp, time(0)+deltatime);
      else        Sched->Schedule(jp);
}

/******************************************************************************/
/*             S e l e c t _ p r o c e s s _ p u t _ f i r s t                */
/******************************************************************************/

int XrdDPMXmi::Select_process_put_first(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error) {
   XrdCmsReq *Request = rp->Request();
   int nbreplies = 0;
   int status;

   struct dpm_putfilestatus *filestatuses=NULL;
   struct dpm_putfilereq reqfiles;
   int overwrite = (pd->opts & XMI_TRUNC) ? 1 : 0;
   char *pp,protocols[32],u_token[256],r_token[CA_MAXDPMTOKENLEN+1];
   time_t retrytime = 0;

   pd->get = false;

   (void)memset((void *)&reqfiles,0,sizeof(struct dpm_putfilereq));
   reqfiles.to_surl = strdup(rp->Path());
   reqfiles.lifetime =        (rp->sopts.have_lifetime) ? rp->sopts.lifetime : Opts.reqput_lifetime;
   reqfiles.f_type =             (rp->sopts.have_ftype) ? rp->sopts.ftype    : Opts.reqput_ftype;
   reqfiles.requested_size =   (rp->sopts.have_reqsize) ? rp->sopts.reqsize  : Opts.reqput_reqsize;
   strcpy(reqfiles.s_token,     (rp->sopts.have_stoken) ? rp->sopts.stoken   : Opts.reqput_stoken);

   strcpy(protocols, XPROTO);                           // protocols had better be large enough
   strcpy(u_token,"Cms XMI-DPM interface put request"); // u_token must be large enough

   error.clear();

   pp = protocols;
   if (dpm_put(1, &reqfiles, 1, &pp, u_token, overwrite, retrytime, r_token, &nbreplies, &filestatuses)<0) {
      status = DPM_FAILED | serrno;
      if (serrno == EINVAL && nbreplies) {
         if (filestatuses[0].errstring != NULL)
            strncpy(pd->fstatbuf,filestatuses[0].errstring,sizeof(pd->fstatbuf));
         pd->fstatbuf[sizeof(pd->fstatbuf)-1] = '\0';
      }
   } else if (nbreplies) {
      status = filestatuses[0].status;
      strcpy(pd->r_token, r_token);
   } else {
      status = DPM_FAILED | EINVAL;
   }

   free(reqfiles.to_surl);
   dpm_free_pfilest(nbreplies, filestatuses);

   if (status != DPM_QUEUED) {
      if ((status&0xf000)==DPM_FAILED && DPMRETRY(status&0xfff)) {
         sendWait(Request, status&0xfff, "Select (for put)");
         return 1;
      }
      sendError(Request, status&0xfff, error.print("during dpm_put"), rp->Path());
      return 1;
   }

   // done for now. Keep request in the queue
   pd->first = false;
   rp->notifyActionTaken(true);
   pd.release();
   return 0;
}

/******************************************************************************/
/*                    S e l e c t _ p r o c e s s _ p u t                     */
/******************************************************************************/

int XrdDPMXmi::Select_process_put(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error) {
   XrdCmsReq *Request = rp->Request();
   int nbreplies = 0;
   int status, retc;
   char *surl, turl[CA_MAXSFNLEN+1];
   turl[0] = '\0';

   // get status for dpm_put request
   struct dpm_putfilestatus *filestatuses=NULL;

   error.clear();

   surl = strdup(rp->Path());
   if (dpm_getstatus_putreq(pd->r_token, 1, &surl, &nbreplies, &filestatuses)<0) {
      status = DPM_FAILED | serrno;
      if (serrno == EINVAL && nbreplies) {
         if (filestatuses[0].errstring != NULL)
            strncpy(pd->fstatbuf,filestatuses[0].errstring,sizeof(pd->fstatbuf));
         pd->fstatbuf[sizeof(pd->fstatbuf)-1] = '\0';
      }
   } else if (nbreplies) {
      status = filestatuses[0].status;
      if (status == DPM_READY && filestatuses[0].turl != NULL) {
         if (strlen(filestatuses[0].turl)>CA_MAXSFNLEN) {
            status = DPM_FAILED | ENAMETOOLONG;
         } else {
            strncpy(turl,filestatuses[0].turl,CA_MAXSFNLEN+1);
            turl[CA_MAXSFNLEN]='\0';
         }
      }
   } else {
      status = DPM_FAILED | EINVAL;
   }

   free(surl);
   dpm_free_pfilest(nbreplies, filestatuses);

   if ((status&0xf000) == DPM_FAILED && DPMRETRY(status&0xfff)) {
      // keep trying, but dont notify of any progress, so eventually the request will be removed
      rp->notifyActionTaken(false);
      pd.release();
      return 0;
   }

   // special case: if we possibly failued due to missing directory elements
   // try to make them.
#ifdef XMI_MAKEP
   if ( ((pd->opts & XMI_MAKEP) || Opts.mkpath_bool) && (pd->opts & (XMI_NEW | XMI_TRUNC)) && status==(DPM_FAILED|ENOENT) && !pd->tried_mkpath ) {
#else
   if ( Opts.mkpath_bool && (pd->opts & (XMI_NEW | XMI_TRUNC)) && status==(DPM_FAILED|ENOENT) && !pd->tried_mkpath ) {
#endif
     if (setupAndUseSession(rp, error)) return 1;
     retc = mkp(rp->Path(),0775);
     if (retc > 0) { // failed to make the path
       if (retc>0 && DPMRETRY(retc)) { // retry if make path failed with a retrable failure
          rp->notifyActionTaken(false);
          pd.release();
          return 0;
       }
       error.clear();
       sendError(Request, retc, error.print("could not create directory path"), rp->Path());
       return 1;
     }
     // path was created or found the exist
     pd->tried_mkpath = true;
     pd->first = true;
     rp->notifyActionTaken(true);
     pd.release();
     return 0;
   }

   if ( (status&0xf000) == DPM_FAILED || (status&0xf000) == DPM_ABORTED ) {
      sendError(Request, status&0xfff, error.print("during dpm_getstatus_putreq"), rp->Path());
      return 1;
   }

   error.clear();

   // now set retc to our return code, 1 to dequeue, 0 to keep in queue
   retc = 0;
   if (status == DPM_READY) {
      sendRedirect(Request, turl, pd->r_token, true);
      retc = 1;
   }

   rp->notifyActionTaken(false);

   if (!retc) {
     // keep in the queue to check again, so dont release private data
     pd.release();
   }
   return retc;
}

/******************************************************************************/
/*             S e l e c t _ p r o c e s s _ g e t _ f i r s t                */
/******************************************************************************/

int XrdDPMXmi::Select_process_get_first(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error) {
   EPNAME("Select_process_get_first");
   XrdCmsReq *Request = rp->Request();
   int nbreplies = 0;
   int status;

   struct dpm_getfilestatus *filestatuses=NULL;
   struct dpm_getfilereq reqfiles;
   time_t retrytime = 0;
   char *pp,protocols[32],u_token[256],r_token[CA_MAXDPMTOKENLEN+1];

   pd->get = true;

   (void)memset((void *)&reqfiles,0,sizeof(struct dpm_getfilereq));
   reqfiles.from_surl = strdup(rp->Path());
   reqfiles.lifetime =       (rp->sopts.have_lifetime) ? rp->sopts.lifetime : Opts.reqget_lifetime;
   reqfiles.f_type =            (rp->sopts.have_ftype) ? rp->sopts.ftype    : Opts.reqget_ftype;
   strcpy(reqfiles.s_token,    (rp->sopts.have_stoken) ? rp->sopts.stoken   : Opts.reqget_stoken);
   reqfiles.flags = 0;
   strcpy(protocols, XPROTO);                           // protocols had better be large enough
   strcpy(u_token,"Cms XMI-DPM interface get request"); // u_token must be large eough

   TRACE(Debug, "issuing dpm get with surl=" << reqfiles.from_surl << " lifetime=" 
      << reqfiles.lifetime << " ftype='" << reqfiles.f_type << "' stoken=" << reqfiles.s_token);

   error.clear();

   pp = protocols;
   if (dpm_get(1, &reqfiles, 1, &pp, u_token,  retrytime, r_token, &nbreplies, &filestatuses)<0) {
      status = DPM_FAILED | serrno;
      if (serrno == EINVAL && nbreplies) {
         if (filestatuses[0].errstring != NULL)
            strncpy(pd->fstatbuf,filestatuses[0].errstring,sizeof(pd->fstatbuf));
         pd->fstatbuf[sizeof(pd->fstatbuf)-1] = '\0';
      }
   } else if (nbreplies) {
      status = filestatuses[0].status;
      strcpy(pd->r_token, r_token);
   } else {
      status = DPM_FAILED | EINVAL;
   }

   free(reqfiles.from_surl);
   dpm_free_gfilest(nbreplies, filestatuses);

   if (status != DPM_QUEUED) {
      if ((status&0xf000)==DPM_FAILED && DPMRETRY(status&0xfff)) {
         sendWait(Request, status&0xfff, "Select (for get)");
         return 1;
      }
      sendError(Request, status&0xfff, error.print("during dpm_get"), rp->Path());
      return 1;
   }

   // done for now. Keep request in the queue
   pd->first = false;
   rp->notifyActionTaken(true);
   pd.release();
   return 0;
}

/******************************************************************************/
/*                    S e l e c t _ p r o c e s s _ g e t                     */
/******************************************************************************/

int XrdDPMXmi::Select_process_get(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error) {
   XrdCmsReq *Request = rp->Request();
   int nbreplies = 0;
   int status, retc;
   char *surl,turl[CA_MAXSFNLEN+1];
   turl[0] = '\0';

   // get status for dpm_get request
   struct dpm_getfilestatus *filestatuses=NULL;

   error.clear();

   surl = strdup(rp->Path());
   if (dpm_getstatus_getreq(pd->r_token, 1, &surl, &nbreplies, &filestatuses)<0) {
      status = DPM_FAILED | serrno;
      if (serrno == EINVAL && nbreplies) {
         if (filestatuses[0].errstring != NULL)
            strncpy(pd->fstatbuf,filestatuses[0].errstring,sizeof(pd->fstatbuf));
         pd->fstatbuf[sizeof(pd->fstatbuf)-1] = '\0';
      }
   } else if (nbreplies) {
      status = filestatuses[0].status;
      if (status == DPM_READY && filestatuses[0].turl != NULL) {
         if (strlen(filestatuses[0].turl)>CA_MAXSFNLEN) {
            status = DPM_FAILED | ENAMETOOLONG;
         } else {
            strncpy(turl,filestatuses[0].turl,CA_MAXSFNLEN+1);
            turl[CA_MAXSFNLEN]='\0';
         }
      }
   } else {
      status = DPM_FAILED | EINVAL;
   }

   free(surl);
   dpm_free_gfilest(nbreplies, filestatuses);

   if ((status&0xf000) == DPM_FAILED && DPMRETRY(status&0xfff)) {
      // keep trying, but dont notify of any progress, so eventually the request will be removed
      rp->notifyActionTaken(false);
      pd.release();
      return 0;
   }

   if ( (status&0xf000) == DPM_FAILED || (status&0xf000) == DPM_ABORTED ) {
      sendError(Request, status&0xfff, error.print("during dpm_getstatus_getreq"), rp->Path());
      return 1;
   }

   error.clear();

   // now set retc to our return code, 1 to dequeue, 0 to keep in queue
   retc = 0;
   if (status == DPM_READY) {
      sendRedirect(Request, turl, pd->r_token, false);
      retc = 1;
   }

   rp->notifyActionTaken(false);
   if (!retc) {
     // keep in the queue to check again, so dont release private data
     pd.release();
   }
   return retc;
}

/******************************************************************************/
/*                          a g e S e s s i o n                               */
/******************************************************************************/

int XrdDPMXmi::ageSession() {
   time_t t = time(NULL);
   return ageSession(t);
}

int XrdDPMXmi::ageSession(const time_t &t) {
   if (m_dpnsSessEst && (m_dpnsSessLastChecked>t || t-m_dpnsSessLastChecked > maxsessage)) {
      dpns_endsess();
      m_dpnsSessEst = false;
   }
   return 0;
}

/******************************************************************************/
/*                       s e t u p S e s s i o n                              */
/******************************************************************************/

int XrdDPMXmi::setupAndUseSession(XrdDPMReq *rp, DPMXmi::XmiErr &err)
{
   int retc;

   err.clear();
   if ((retc=setupAndUseSession())) {
      if (DPMRETRY(retc)) { sendWait(rp->Request(), retc, err.print(0)); }
      else { sendError(rp->Request(), retc, err.print(0), rp->Path()); }
   }
   err.clear();

   return retc ? 1 : 0;
}

int XrdDPMXmi::setupAndUseSession()
{
   time_t t = time(NULL);
   ageSession(t);

   if (!*m_ServerAndPort) return EHOSTUNREACH;

   if (!m_dpnsSessEst) {
      std::ostringstream o;
      std::string s;
      o << XrdDPMXmiCVSID;
#if defined(XRDDPM_GCC_VERSION)
      o << ": GCC version " << XRDDPM_GCC_VERSION;
#endif
      o << ": XrdDPM version " << XrdDPMVERSIONSTR;
#if defined (XrdCompileVERSION)
      o << ": Xrootd version " << XrdCompileVERSION;
#endif
#if defined (XRDDPM_BUILDDATE)
      o << ": Built " << XRDDPM_BUILDDATE;
#endif
      s = o.str();
      if (dpns_startsess(m_ServerAndPort, (char *)s.c_str())<0) {
         return serrno;
      }
      m_dpnsSessEst = true;
   }

   if (m_dpnsSessEst)
      m_dpnsSessLastChecked = t;

   return 0;
}

/******************************************************************************/
/*                       c l o s e S e s s i o n                              */
/******************************************************************************/

int XrdDPMXmi::closeSession()
{
   if (m_dpnsSessEst) {
      dpns_endsess();
      m_dpnsSessEst = false;
   }

   return 0;
}

/******************************************************************************/
/*                                  m k p                                     */
/******************************************************************************/

int XrdDPMXmi::mkp(const char *path, const mode_t &mode) const
{
   char *next_path, this_path[CA_MAXPATHLEN+1];
   struct dpns_filestatg dpnsbuf;
   int retc;

   if (strlen(path)>CA_MAXPATHLEN) return ENAMETOOLONG;
   strcpy(this_path, path);

// Trim off the trailing slash so that we make everything but the last component
//
   if (!(retc = strlen(this_path))) {
      return ENOENT;
   }
   while(retc && this_path[retc-1] == '/')
        {retc--; this_path[retc] = '\0';}

// Typically, the path exists. So, do a quick check before launching into it
//
   if (!(next_path = rindex(this_path, (int)'/'))
   ||  next_path == this_path) {
      return 0;
   }
   *next_path = '\0';
   if (!dpns_statg(this_path, NULL, &dpnsbuf)) {
      return 0;
   }
   *next_path = '/';

// Start creating directories starting with the root
//
   next_path = this_path;
   while((next_path = index(next_path+1, int('/'))))
        {if (next_path[1] == '/') continue;
         *next_path = '\0';
         if (dpns_statg(this_path, NULL, &dpnsbuf)<0) {
           if (serrno == ENOENT) {
             if ( dpns_mkdir(this_path, mode)<0 && serrno != EEXIST ) {
               return serrno;
             }
           } else {
             return serrno;
           }
         }
         *next_path = '/';
        }

// All done
//
   return -1;
}

/******************************************************************************/
/*           S e t D e f a u l t s A n d R e a d C o n f i g                  */
/******************************************************************************/

int XrdDPMXmi::SetDefaultsAndReadConfig(XrdCmsXmiEnv *Env)
{
   EPNAME("SetDefaultsAndReadConfig");
   const char *var, *val, *ConfigFN = Env->ConfigFN;
   int  cfgFD, retc, NoGo = 0;
   int  v;
   XrdOucEnv myEnv;
   XrdOucStream CFile(eDest, "anon@localhost", &myEnv, "=====> ");

// Set defaults
   Opts.dpns_host[0]     = '\0';
   Opts.dpns_port        =    0;
   Opts.xrd_server_port  =    0;
   Opts.mkpath_bool      =false;
   Opts.reqput_lifetime  =    0;
   Opts.reqput_ftype     =  'P';
   Opts.reqput_stoken[0] = '\0';
   Opts.reqput_reqsize   =    0;
   Opts.reqget_lifetime  =    0;
   Opts.reqget_ftype     = '\0';
   Opts.reqget_stoken[0] = '\0';
   Opts.principal[0]     = '\0';
   Opts.voname[0]        = '\0';
   Opts.nbfqans          =    0;
   Opts.authopt          =    0;
   Opts.authlib[0]       = '\0';
   Opts.authparm[0]      = '\0';
   Opts.authorization    = (XrdAccAuthorize *)0;

// set default for dpns_host
//
   val = getenv("DPNS_HOST");
   if (val && strlen(val)<=CA_MAXHOSTNAMELEN)
      strcpy(Opts.dpns_host, val);

// set default xrootd server port
// nb atoi not thread safe
   val = getenv("XROOTD_SERVER_PORT");
   if (val)
      Opts.xrd_server_port = atoi(val);

// Try to open the configuration file.
//
   if ( (cfgFD = open(ConfigFN, O_RDONLY, 0)) < 0)
      {eDest->Emsg("Config", errno, "open config file", ConfigFN);
       return 1;
      }
   CFile.Attach(cfgFD);
// Now start reading records until eof.
//
   while((var = CFile.GetMyFirstWord())) {
      if (!strncmp(var, "dpm.", 4)) {
         var += 4;

         if (!strcmp("xmiauthlib",var)) {
            if (!(val = CFile.GetWord()) || strlen(val)>=sizeof(Opts.authlib)) {
               eDest->Emsg("Config", "xmiauthlib not specified or invalid name.");NoGo |= 1;}
            else { 
               strcpy(Opts.authlib, val);
               if (!CFile.GetRest(Opts.authparm, sizeof(Opts.authparm))) {
                  eDest->Emsg("Config", "xmiauthlib parameters invalid."); NoGo |= 1;}
            }
         }

         if (!strcmp("xmiselectauthorize",var)) {
            Opts.authopt = 1;
         }

         if (!strcmp("putlifetime",var)) {
            if (!(val = CFile.GetWord())) {
               eDest->Emsg("Config", "'put' lifetime value not specified."); NoGo |= 1;
            } else { if(XrdOuca2x::a2tm(*eDest, "putlifetime value", val, &v))
                      NoGo |= 1; else Opts.reqput_lifetime = v;}
         }

         if (!strcmp("putfiletype",var)) {
            if (!(val = CFile.GetWord()) || strlen(val)!=1) {
               eDest->Emsg("Config", "'put' file type value not specified or invalid.");NoGo |= 1;}
            else { Opts.reqput_ftype = (*val == '-') ? '\0' : *val; }
         }

         if (!strcmp("putstoken",var)) {
            if (!(val = CFile.GetWord()) || strlen(val)>CA_MAXDPMTOKENLEN) {
               eDest->Emsg("Config", "'put' space token not specified or invalid.");NoGo |= 1;}
            else { strcpy(Opts.reqput_stoken,val); }
         }

         if (!strcmp("surlprefix",var)) {
            if (!(val = CFile.GetWord()) || strlen(val)>=sizeof(l_madpfx)) {
               eDest->Emsg("Config", "'surlprefix' not specified or invalid.");NoGo |= 1;}
            else { 
               if (l_madpfxlen && strcmp(val, l_madpfx)) {
                  eDest->Emsg("Config", "surl prefix already set to a different value.");NoGo |= 1;}
                  else {strcpy(l_madpfx,val);l_madpfxlen=strlen(val);}
            }
         }

         if (!strcmp("putsize",var)) {
            if (!(val = CFile.GetWord())) {
               eDest->Emsg("Config", "defult 'put' size not specified.");NoGo |= 1;
            } else { if (XrdOuca2x::a2i(*eDest, "putsize value", val, &v))
                      NoGo |= 1; else Opts.reqput_reqsize = v;}
         }

         if (!strcmp("getlifetime",var)) {
            if (!(val = CFile.GetWord())) {
               eDest->Emsg("Config", "'get' lifetime value not specified.");NoGo |= 1;
            } else { if (XrdOuca2x::a2tm(*eDest, "getlifetime value", val, &v))
                      NoGo |= 1; else Opts.reqget_lifetime = v;}
         }

         if (!strcmp("getfiletype",var)) {
            if (!(val = CFile.GetWord()) || strlen(val)!=1) {
               eDest->Emsg("Config", "'get' file type value not specified or invalid.");NoGo |= 1;}
            else { Opts.reqget_ftype = (*val == '-') ? '\0' : *val; }
         }

         if (!strcmp("getstoken",var)) {
            if (!(val = CFile.GetWord()) || strlen(val)>CA_MAXDPMTOKENLEN) {
               eDest->Emsg("Config", "'get' space token not specified or invalid.");NoGo |= 1;}
            else { strcpy(Opts.reqget_stoken,val); }
         }

         if (!strcmp("dpnshost",var)) {
             if (!(val = CFile.GetWord()) || strlen(val)>CA_MAXHOSTNAMELEN) {
                 eDest->Emsg("Config", "'dpnshost' has zero length or exceeds maximum length.");NoGo |= 1;}
            else { strcpy(Opts.dpns_host,val);}
         }

         if (!strcmp("dpnsport",var)) {
             if (!(val = CFile.GetWord())) {
                 eDest->Emsg("Config", "'dpnsport' is not specified.");NoGo |= 1;
             } else { if (XrdOuca2x::a2i(*eDest, "dpnsport value", val, &v))
                          NoGo |= 1; else Opts.dpns_port = v;}
         }
	 
         if (!strcmp("xrdserverport",var)) {
	   if (!(val = CFile.GetWord())) {
                 eDest->Emsg("Config", "'xrdserverport' is not specified.");NoGo |= 1;
              } else { if (XrdOuca2x::a2i(*eDest, "xrdserverport value", val, &v))
		NoGo |= 1; else Opts.xrd_server_port = v;}
         }

         if (!strcmp("mkpath",var)) {
           if (!(val = CFile.GetWord())) {
             eDest->Emsg("Config", "'mkpath' is not specified.");NoGo |= 1;
           } else { if (XrdOuca2x::a2i(*eDest, "mkpath value", val, &v))
             NoGo |= 1; else Opts.mkpath_bool = v;}
         }
      }
   }

// Now check if any errors occured during file i/o
//
   if ((retc = CFile.LastError()))
      NoGo = eDest->Emsg("Config", retc, "read config file", ConfigFN);
   CFile.Close();

// Try to open the configuration file again for principal/voname/fqans scan.
//
   if ( (cfgFD = open(ConfigFN, O_RDONLY, 0)) < 0)
      {eDest->Emsg("Config", errno, "open config file", ConfigFN);
       return 1;
      }
   CFile.Attach(cfgFD);
   while((var = CFile.GetLine())) {
      while(*var && *var==' ') var++;
      if (*var=='#') continue;
      if (!strncmp("dpm.principal ",var,14)) {
         var += 14;
         while(*var && *var==' ') var++;
         const char *tmp = var;
         while(*tmp) tmp++;
         while(tmp!=var && *(tmp-1)==' ') tmp--;
         size_t s = tmp-var;
         if (s>CA_MAXCSECNAMELEN) {eDest->Emsg("Config","invalid argument for principal name."); NoGo |= 1;}
         else {strncpy(Opts.principal,var,s);Opts.principal[s]='\0';}
      }
      if (!strncmp("dpm.voname ",var,11)) {
         var += 11;
         while(*var && *var==' ') var++;
         const char *tmp = var;
         while(*tmp) tmp++;
         while(tmp!=var && *(tmp-1)==' ') tmp--;
         size_t s = tmp-var;
         if (s>CA_MAXCSECNAMELEN) {eDest->Emsg("Config","invalid argument for voname name."); NoGo |= 1;}
         else {strncpy(Opts.voname,var,s);Opts.voname[s]='\0';}
      }
      if (!strncmp("dpm.fqan ",var,9)) {
         var += 9;
         while(*var && *var==' ') var++;
         const char *tmp = var;
         while(*tmp) tmp++;
         while(tmp!=var && *(tmp-1)==' ') tmp--;
         size_t s = tmp-var;
         if (Opts.nbfqans >= sizeof(Opts.fqans)/(CA_MAXCSECNAMELEN+1)) {eDest->Emsg("Config","too many fqans."); NoGo |= 1;}
         else if (s>CA_MAXCSECNAMELEN) {eDest->Emsg("Config","invalid argument for fqan name."); NoGo |= 1;}
         else {strncpy(Opts.fqans[Opts.nbfqans],var,s);Opts.fqans[Opts.nbfqans][s]='\0';Opts.nbfqans++;}
      }
   }

// Now check if any errors occured during file i/o
//
   if ((retc = CFile.LastError()))
      NoGo = eDest->Emsg("Config", retc, "read config file", ConfigFN);
   CFile.Close();

   DEBUG("Opts.dpns_host        = "<<Opts.dpns_host);
   DEBUG("Opts.dpns_port        = "<<Opts.dpns_port);
   DEBUG("Opts.xrd_server_port  = "<<Opts.xrd_server_port);
   DEBUG("Opts.mkpath_bool      = "<<Opts.mkpath_bool);
   DEBUG("Opts.reqput_lifetime  = "<<Opts.reqput_lifetime);
   DEBUG("Opts.reqput_ftype     = "<<Opts.reqput_ftype);
   DEBUG("Opts.reqput_stoken    = "<<Opts.reqput_stoken);
   DEBUG("Opts.Reqput_reqsize   = "<<Opts.reqput_reqsize);
   DEBUG("Opts.reqget_lifetime  = "<<Opts.reqget_lifetime);
   DEBUG("Opts.reqget_ftype     = "<<Opts.reqget_ftype);
   DEBUG("Opts.reqget_stoken    = "<<Opts.reqget_stoken);
   DEBUG("Opts.principal        = "<<Opts.principal);

// Return final return code
//
   return NoGo;
}

/******************************************************************************/
/*                              r c T o S t r                                 */
/******************************************************************************/

const char *XrdDPMXmi::rcToStr(const int &rc) const
{
   const char *ecode;

// Convert errnumber to err string
//
   switch(rc)
         {case ENOENT:        ecode = "ENOENT";      break;
          case EPERM:         ecode = "EISDIR";      break;// Wierd but documented
          case EEXIST:        ecode = "ENOTEMPTY";   break;// Wierd but documented
          case EACCES:        ecode = "EACCES";      break;
          case ENOSPC:        ecode = "ENOSPC";      break;
          case EFAULT:        ecode = "ENOMEM";      break;// No other choice here
          case ENOTDIR:       ecode = "ENOTDIR";     break;
          case ENAMETOOLONG:  ecode = "ENAMETOOLONG";break;
          case SENOSHOST:     ecode = "ENETUNREACH"; break;
          case SENOSSERV:     ecode = "ENETUNREACH"; break;
          case SECOMERR:      ecode = "ECOMM";       break;// Comm error
          default:            ecode = "EINVAL";
                              break;
         }

   return ecode;
}

/******************************************************************************/
/*                             s e n d E r r o r                              */
/******************************************************************************/
  
int XrdDPMXmi::sendError(XrdCmsReq *reqP, const int &rc, const char *opn, 
                                                  const char *path) const
{
   EPNAME("sendError");
   const char *ecode;
   char buff[256];
   int n;

// Convert errnumber to err string
//
   ecode = rcToStr(rc);

// send error
//
   n = snprintf(buff, sizeof(buff), "%s failed; %s", opn, sstrerror(rc));
   reqP->Reply_Error(ecode, buff);
   DEBUG("msg='" <<buff <<"' path=" <<path);
   return 1;
}

/******************************************************************************/
/*                              s e n d W a i t                               */
/******************************************************************************/

int XrdDPMXmi::sendWait(XrdCmsReq *reqP, const int &rc, const char *opn) const
{
   static int Retries = 255;

// we will tell the client to try again later
//
   Retries++;
   if (!(Retries & 255)) eDest->Emsg("dpm", opn, "failed;", sstrerror(rc));
   reqP->Reply_Wait(retryTime);
   return 1;
}

/******************************************************************************/
/*                          s e n d R e d i r e c t                           */
/******************************************************************************/
  
void XrdDPMXmi::sendRedirect(XrdCmsReq *Request,
                             const char *turl,
                             const char *token,
                             const bool &put) const
{
   EPNAME("sendRedirect");
   char sfn[CA_MAXHOSTNAMELEN+CA_MAXSFNLEN+8];
   char diskserver[CA_MAXHOSTNAMELEN+7];
   char *buff=NULL;
   char *p,*p2;
   int NoGo = 0;

// expect a turl format like:
//      xroot://server.com/server2.com:/path
// or   xroot://server2.com:/path
//
// in the first case use 'server.com' as the redirection address as 'server2.com:/path' as sfn
// in the second case use 'server2.com' as the redirection address and '/path' as sfn

// find first ":/"
   p = strstr(turl,":/");

// skip any more '/'
   if (p) p += 2;
   while(p && *p=='/') p++;
   p2 = p;

// find next '/'
   while(p2 && *p2 && *p2!='/') p2++;

// backup over ":" if present
   if (p2 && *(p2-1)==':') --p2;

// copy string between as the disk server name
   if (p) {
     size_t s = p2-p;
     if ((s+6) < sizeof(diskserver)) {
        strncpy(diskserver, p, s);
        if (Opts.xrd_server_port != 0) {
            sprintf(diskserver+s,":%hu",Opts.xrd_server_port);
        } else {
            diskserver[s] = '\0';
        }
     } else { NoGo |= 1; }
   }

// copy remainer after the disk server into sfn
   if (p2) {
      while(*p2==':' || *p2=='/') p2++;
      while(*p) p++;
      size_t s = p-p2;
      if (s < sizeof(sfn)) {
         strncpy(sfn, p2, s);
         sfn[s] = '\0';
      } else { NoGo |= 1; }
   }

// note if there were processing problems
   if (!p || !strlen(diskserver) || !strlen(sfn)) NoGo |= 1;

// check if the token is too long
   if (strlen(token)>CA_MAXDPMTOKENLEN) NoGo |= 1;

   size_t needed;
   if (!NoGo) {
      // size needed for send buffer, is for &dpm.sfn=...&dpm.tok=[g|p]...
      //
      needed = strlen(sfn) + strlen(token) + 20;
      //
      // allow for escaping the sfn for various characters
      int num_esc = 0;
      p = sfn;
      while(*p) { if (ESCAPEURLCHAR(*p)) num_esc++; p++; }
      needed += 2*num_esc;
      buff = (char *)malloc(needed);
      if (!buff) NoGo |= 1;
   }

   if (!NoGo) {
      strcpy(buff, "&dpm.sfn=");
      p = buff; while(*++p);
      p2 = sfn;
      while(*p2) {
         if (!ESCAPEURLCHAR(*p2)) *p++ = *p2;
         else { unsigned short c = *p2; *p++ = '%'; snprintf(p, 3, "%02hx", c); p+=2; } 
         p2++;
      }
      *p = '\0';
      sprintf(&buff[strlen(buff)],"&dpm.tok=%c%s",put ? 'p' : 'g',token);
      if (strlen(buff)+1 != needed) NoGo |= 1;
   }

   if (!NoGo) {
      TRACE(Debug, "sending redirection to disk server " << diskserver << " cgi info " << buff);
      Request->Reply_Redirect(diskserver,(const char *)0,buff);
   } else {
      TRACE(Debug, "sending error: EINVAL invalid turl");
      sendError(Request, EINVAL, "Invalid turl", turl);
   }
   if (buff) free(buff);
}

/******************************************************************************/
/*                          n o t S u p p o r t e d                           */
/******************************************************************************/
  
int XrdDPMXmi::notSupported(XrdCmsReq *rp, const char *opn, const char *path) const
{
   char buff[2048];
   int n;

   n = sprintf(buff, "Unable to %s %s; operation not supported.",opn,path);
   rp->Reply_Error(buff);
   return 1;
}

/******************************************************************************/
/*                           p r e f i x _ s u r l                            */
/******************************************************************************/

static char *prefix_mad(const char *surl, char *out, size_t out_len)
{
   char *p = out;
   int pathlen = strlen(surl);
   int addslash = (*surl != '/');
   if ((l_madpfxlen + addslash + pathlen) >= out_len) return NULL;
   if (l_madpfxlen) {strcpy(out, l_madpfx); out += l_madpfxlen;}
   if (addslash) {*out='/';out++;}
   strcpy(out, surl);
   return p;
}

/******************************************************************************/
/*                      X r d O u c N a m e 2 N a m e                         */
/******************************************************************************/
extern "C" 
{
XrdOucName2Name *XrdOucgetName2Name(XrdOucgetName2NameArgs)
{
   XrdDPMXmiN2N *inst = new XrdDPMXmiN2N(eDest, lroot, rroot);

  if (parms) eDest->Say("++++++ XrdDPMXmiN2N initializing. Local lfn prefix '",parms,"'.");
   else
     eDest->Say("++++++ XrdDPMXmiN2N initializing. Local lfn prefix is null");

   if (inst->set_mad(parms)) {
      eDest->Say("++++++ XrdDPMXmiN2N failed while setting prefix");
      delete inst;
      inst = NULL;
   }

   return (XrdOucName2Name *)inst;
}
}

/******************************************************************************/
/*                           C o n s t r u c t o r                            */
/******************************************************************************/

XrdDPMXmiN2N::XrdDPMXmiN2N(XrdSysError *erp, const char *lpfx, const char *rpfx)
{
   eDest = erp;
}

/******************************************************************************/
/*                             l f n 2 p f n                                  */
/******************************************************************************/
int XrdDPMXmiN2N::lfn2pfn(const char *lfn, char *buff, int blen)
{
   return eDest->Emsg("XrdDPMXmiN2N::lfn2pfn",-EOPNOTSUPP,"generate physical path lfn=",lfn);
}

/******************************************************************************/
/*                             p f n 2 l f n                                  */
/******************************************************************************/
int XrdDPMXmiN2N::pfn2lfn(const char *pfn, char *buff, int blen)
{
   return eDest->Emsg("XrdDPMXmiN2N::pfn2lfn",-EOPNOTSUPP,"generate logical path pfn=",pfn);
}

/******************************************************************************/
/*                             l f n 2 r f n                                  */
/******************************************************************************/
int XrdDPMXmiN2N::lfn2rfn(const char *lfn, char *buff, int blen)
{
   return XrdDPMXmi::defaultLfn2rfn(lfn, buff, blen);
}

/******************************************************************************/
/*                              s e t _ m a d                                 */
/******************************************************************************/

int XrdDPMXmiN2N::set_mad(const char *mad)
{
  if (l_madpfxlen && (!mad || (*mad && strcmp(mad, l_madpfx)))) {
     eDest->Say("Attempting to set different local lfn prefix to that which is already set.");
     return 1;
   }

   // keep copy in global for use by DPM Xmi
   //
   if (mad && *mad && strlen(mad)<sizeof(l_madpfx)) {
     strcpy(l_madpfx, mad);
     l_madpfxlen = strlen(mad);
   }
   return 0;
}
