/******************************************************************************/
/*                                                                            */
/*               X r d D P M X m i Q u e u e A n d R e q . c c                */
/*                                                                            */
/* DPM handling and various by David Smith, CERN                              */            
/* Other parts:                                                               */
/*                                                                            */
/* (c) 2007 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

const char XrdDPMQueueCVSID[] = "$Id: XrdDPMXmiQueueAndReq.cc,v 1.4 2009/10/05 09:33:37 dhsmith Exp $";

#include <stdlib.h>
#include <string.h>

#include "XrdDPM/XrdDPMXmiQueueAndReq.hh"
#include "XrdCms/XrdCmsXmi.hh"
#include "XrdSys/XrdSysPlatform.hh"

#include "XrdDPM/XrdDPMXmi.hh"
#include "XrdDPM/XrdDPMXmiTrace.hh"

#define INITIAL_PAUSETIME 1

/******************************************************************************/
/*                  X r d D P M Q u e u e   m e m b e r s                     */
/******************************************************************************/
/******************************************************************************/
/*                            C o n s t r u c t o r                           */
/******************************************************************************/

XrdDPMQueue::XrdDPMQueue() : queueMutex(0)
{
   nextFree = 0;
   firstQueue = 0;
   lastQueue = 0;
   numFree = 0;
   numinQ = 0;
   isWaiting = 0;
   N2N = 0;
}

/******************************************************************************/
/*                                 A l l o c                                  */
/******************************************************************************/
  
XrdDPMReq *XrdDPMQueue::Alloc(const XrdCmsXmiReq::ReqType rtype, 
                              XrdCmsReq                  *reqP,
                              const char                 *Path,
                              void                       *pd)
{
   XrdDPMReq *rp;
   int rc;

// Allocate a request object. If we have no memory, tell the requester
// to try again in a minute.
//
   queueMutex.Lock();
   if (nextFree) {
      rp = nextFree;
      nextFree = nextFree->Next;
      numFree--;
   } else if (!(rp = new XrdDPMReq())) {
      queueMutex.UnLock();
      reqP->Reply_Wait(retryTime);
      return (XrdDPMReq *)0;
   }
   queueMutex.UnLock();

// Initialize the path which may be a mapping from the lfn to the MSS name
//
   if (N2N) rc = N2N->lfn2rfn(Path, rp->thePath, XrdDPMReq::PathSize);
      else rc = XrdDPMXmi::defaultLfn2rfn(Path, rp->thePath, XrdDPMReq::PathSize);

// Make sure the initialization succeeded
//
   if (rc) {
      const char *err = "Unable to map lfn to DPM name.";
      reqP->Reply_Error(err);
      eDest->Emsg("DPMReq", rc, "map lfn to DPM name.");
      rp->Recycle();
      return 0;
   }

   XrdCmsReq *reqPcopy = new XrdCmsReq(*reqP);

// Initalise the request object
//
   rp->init(reqPcopy, rtype, this, pd);

// Get any extra options: put them in the request in the sopts struct
//
   myXmi->getExtraSurlOptions(Path, rp);

// Return the request
//
   return rp;
}

/******************************************************************************/
/*                                   S e t                                    */
/******************************************************************************/
  
void XrdDPMQueue::Set(XrdDPMXmi *xmi, XrdCmsXmiEnv *Env)
{

// Copy out needed stuff from the environment
//
   eDest = Env->eDest;         // -> Error message handler
   Trace = Env->Trace;         // -> Trace handler
   N2N   = Env->Name2Name;     // -> lfn mapper
   myXmi = xmi;                // DPM xmi object
}

/******************************************************************************/
/*                           M a k e O n e P o l l                            */
/******************************************************************************/

int XrdDPMQueue::MakeOnePoll(int &nqueued)
{
   XrdDPMReq *rp,*p = 0;
   time_t st,t=0,mint=-1;
   int ret;

   queueMutex.Lock();
   rp = firstQueue;
   st = time(NULL);

   nqueued = numinQ;

   while(rp) {
      if (rp->lastPollStarted > st) {
         p = rp;
         break;
      } else {
         // rage: minimum age for request of this state
         // cage: current age of request
         // remain: time remaining until current request is eligable
         //
         int rage;
         if (rp->state == XrdDPMReq::never_called) rage = 0;
         else rage = rp->pauseTime;
         int cage = st - rp->lastPollStarted;
         int remain = rage>cage ? rage-cage : 0;

         // keep track of smallest time until there is an eligable request
         //
         if (mint == -1 || remain<mint)
            mint = remain;

         // now choose this entry if it is eligable
         if (!remain && (t==0 || t > rp->lastPollStarted)) {
            p = rp;
            t = rp->lastPollStarted;
         }
      }
      rp = rp->Next;
   }
   queueMutex.UnLock();

   if (!p) {
      // nothing avilable or eligable: return (-1) nothing available
      // else the minimum number of seconds until something is available
      return mint;
   }

   p->lastPollStarted = st;
   if (p->state == XrdDPMReq::never_called) {
      p->lastProgressMade = st;
      p->state = XrdDPMReq::first_action;
      p->pauseTime = INITIAL_PAUSETIME;
   }

   ret = do_poll_switch(p, false);

   if (ret) {
     // done
     Remove(p);
     p->Recycle();
   } else {
      st = time(NULL);
      if (p->lastProgressMade+maxqueueage < st) {
         // too long since the entry made progress: remove it
         do_poll_switch(p, true);
         Remove(p);
         p->Recycle();
      }
   }

   return 0;
}

/******************************************************************************/
/*                                W a i t 4 Q                                 */
/******************************************************************************/

int XrdDPMQueue::Wait4Q(const int &prevnum, const int &max)
{
   int temp=0;

// When something is added to the queue return(0)
// While there is something existing in the queue
//   return(non-0) every max seconds
//
   queueMutex.Lock();
   if (numinQ == prevnum) {
      isWaiting = 1;
      do {
         temp = queueMutex.Wait(max);
       } while(!temp && isWaiting == 1);
      temp = isWaiting;
      isWaiting = 0;
   }
   queueMutex.UnLock();
   return temp;
}

/******************************************************************************/
/*                     d o _ p o l l _ s w i t c h                            */
/******************************************************************************/

int XrdDPMQueue::do_poll_switch(XrdDPMReq *p, const bool &destruct)
{
   int ret;

   switch(p->cmsXmiRtype) {
      case XrdCmsXmiReq::do_chmod:
         ret = myXmi->Chmod_process(p, destruct);
         break;

      case XrdCmsXmiReq::do_mkdir:
         ret = myXmi->Mkdir_process(p, destruct);
         break;

      case XrdCmsXmiReq::do_mkpath:
         ret = myXmi->Mkpath_process(p, destruct);
         break;  

      case XrdCmsXmiReq::do_mv:
         ret = myXmi->Rename_process(p, destruct);
         break;

      case XrdCmsXmiReq::do_rm:
         ret = myXmi->Remove_process(p, destruct);
         break;

      case XrdCmsXmiReq::do_rmdir:
         ret = myXmi->Remdir_process(p, destruct);
         break;

      case XrdCmsXmiReq::do_stage:
         ret = myXmi->Select_process(p, destruct);
         break;

      case XrdCmsXmiReq::do_stat:
         ret = myXmi->Stat_process(p, destruct);
         break;

      default:
         // of the possible enumerated ReqTypes the follow are not covered
         // do_prep - returns not supported immediatly
         // possibly load and space in the future:
         //   currently not supported or enabled in the plugin via XeqMode() settings
         ret = 1;
         break;
   }

   return ret;
}

/******************************************************************************/
/*                                R e m o v e                                 */
/******************************************************************************/
  
void XrdDPMQueue::Remove(const XrdDPMReq *toDel)
{
   XrdDPMReq *rp;

   queueMutex.Lock();
   rp = toDel->Prev;
   if (rp) {
      rp->Next = toDel->Next;
   } else {
      firstQueue = toDel->Next;
   }
   rp = toDel->Next;
   if (rp) {
     rp->Prev = toDel->Prev;
   } else {
     lastQueue = toDel->Prev;
   }
   numinQ--;
   queueMutex.UnLock();

   return;
}

/******************************************************************************/
/*                     X r d D P M R e q   m e m b e r s                      */
/******************************************************************************/
/******************************************************************************/
/*                                  i n i t                                   */
/******************************************************************************/

void XrdDPMReq::init(XrdCmsReq *reqP, XrdCmsXmiReq::ReqType rtype, XrdDPMQueue *queue, void *pd)
{
   this->cmsReq  = reqP;
   this->cmsXmiRtype = rtype;
   this->myQueue = queue;
   this->pd = pd;

   this->Next    = 0;
   this->Prev    = 0;
   this->state = XrdDPMReq::never_called;

   this->sopts.have_stoken = false;
   this->sopts.have_ftype = false;
   this->sopts.have_lifetime = false;
   this->sopts.have_reqsize = false;
}

/******************************************************************************/
/*                                 Q u e u e                                  */
/******************************************************************************/
  
void XrdDPMReq::Queue()
{
// Add this request to the slot table
//
   myQueue->queueMutex.Lock();
   if (!myQueue->firstQueue) {
      myQueue->firstQueue = this;
      myQueue->lastQueue = this;
   } else {
      myQueue->lastQueue->Next = this;
      this->Prev = myQueue->lastQueue;
      myQueue->lastQueue = this;
   }

   myQueue->numinQ++;

   if (myQueue->isWaiting) {
      myQueue->isWaiting = 0;
      myQueue->queueMutex.Signal();
   }

   // mark the queue time
   //
   lastPollStarted = time(NULL); 
   myQueue->queueMutex.UnLock();
}

/******************************************************************************/
/*                               R e c y c l e                                */
/******************************************************************************/
  
void XrdDPMReq::Recycle()
{
   if (cmsReq) { delete cmsReq; cmsReq = 0; }
// Put this object on the free queue or delete it
//
   myQueue->queueMutex.Lock();
   if (myQueue->numFree >= myQueue->maxFree) {
      myQueue->queueMutex.UnLock();
      delete this;
   } else {
      this->Prev = 0;
      this->Next = myQueue->nextFree;
      myQueue->nextFree = this;
      myQueue->numFree++;
      myQueue->queueMutex.UnLock();
   }
   return;
}

/******************************************************************************/
/*                   n o t i f y A c t i o n T a k e n                        */
/******************************************************************************/

void XrdDPMReq::notifyActionTaken(const bool &madeProgress)
{
   state = waiting_action;

   if (madeProgress) {
      time_t t = time(NULL);
      lastProgressMade = t;
      pauseTime = INITIAL_PAUSETIME;
   } else {
      if (pauseTime == 0) pauseTime = 1;
      else pauseTime *= 2;
      if (pauseTime > 600) pauseTime = 600;
   }
}
